(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Substitution imports Term begin


(* unifiers and properties of unifiers *)
definition unifiers where "unifiers pairs = {\<sigma>. \<forall> p \<in> pairs. fst p \<cdot> \<sigma> = snd p \<cdot> \<sigma>}"

definition isMgu where "isMgu \<sigma> pairs == (\<sigma> \<in> unifiers pairs \<and> (\<forall> \<tau> \<in> unifiers pairs. (\<exists> \<gamma> . \<tau> = \<sigma> \<circ>s \<gamma>)))"

definition isSMgu where "isSMgu \<sigma> pairs == (\<sigma> \<in> unifiers pairs \<and> (\<forall> \<tau> \<in> unifiers pairs. \<tau> = \<sigma> \<circ>s \<tau>))"

lemma "isSMgu \<sigma> pairs \<Longrightarrow> isMgu \<sigma> pairs" by (unfold isMgu_def isSMgu_def, auto)


lemma unifiersInter[simp] : "unifiers (s \<union> t) = unifiers s \<inter> unifiers t"
  by (unfold unifiers_def, blast)

lemma unifiersInsert : "unifiers (insert p t) = {\<sigma> . fst p\<cdot>\<sigma> = snd p\<cdot>\<sigma>} \<inter> unifiers t" (* simp not added for readability *)
  by (unfold unifiers_def, auto)

lemma unifiersSwap : "unifiers (insert (s,t) pairs) = unifiers (insert (t,s) pairs)"
  by (auto simp: unifiersInsert)

lemma unifiersSwapDirect[simp] : "unifiers (insert (Fun f ss, Var x) pairs) = unifiers (insert (Var x, Fun f ss) pairs)"
  by (simp add: unifiersSwap)

lemma unifiersEmpty[simp] : "unifiers {} = UNIV"
  by (unfold unifiers_def, auto)

lemma unifiersIdent[simp] : "unifiers (insert (t,t) pairs) = unifiers pairs" 
  by (auto simp: unifiersInsert)

lemma unifiersDrop[simp] : fixes s :: "('f,'v) term" and zs :: "(('f,'v) term \<times> ('f,'v) term) set"  
  assumes "(s, t) \<in> zs" shows "unifiers {(s,t)} \<inter> unifiers zs = unifiers zs"
using assms
by (unfold unifiers_def, auto)

lemma unifiersElem : fixes zs :: "(('f,'v) term \<times> ('f,'v) term) set"  
  assumes "\<sigma> \<in> unifiers zs" and "pair \<in> zs" shows "fst pair\<cdot>\<sigma> = snd pair\<cdot>\<sigma>"
using assms
by (unfold unifiers_def, auto)

lemma unifiersOccur[simp] : fixes s :: "('f,'v) term" and pairs :: "(('f,'v) term \<times> ('f,'v) term) set"  
  assumes tx: "t \<noteq> Var x" and xt: "x \<in> vars_term t" shows "unifiers (insert (Var x,t) pairs) = {}"
using assms
proof -
  from xt have "t \<unrhd> Var x" by (rule supteqp_var) 
  hence "t \<rhd> Var x \<or> t = Var x" by (simp only: supteqp_iff_suptp_or_eq)
  with tx have subt: "t \<rhd> Var x" by simp
  show ?thesis 
  proof (unfold unifiers_def, rule set_ext, auto)
    fix \<sigma> 
    assume eq: "get_sub \<sigma> x = t \<cdot> \<sigma>"
    from subt have "t \<cdot> \<sigma> \<rhd> Var x \<cdot> \<sigma>" by (rule suptp_stable)
    hence "t \<cdot> \<sigma> \<noteq> Var x \<cdot> \<sigma>" by (rule  suptp_imp_neq)
    with eq have False by simp 
    thus "\<exists>p\<in>pairs. fst p \<cdot> \<sigma> \<noteq> snd p \<cdot> \<sigma>" ..
  qed
qed


(* more properties of substitutions *)

lemma substDom[simp] : "dom {x\<leftarrow>t} = (if (t = Var x) then {} else {x})"
proof (cases "{x\<leftarrow>t}")
  case (Sub sigma)
  have "dom (Sub sigma) = (if t = Var x then {} else {x})" 
  proof (rule set_ext, simp only: dom.simps)
    fix y 
    from Sub show "(y \<in> {x. sigma x \<noteq> Var x}) = (y \<in> (if t = Var x then {} else {x}))"
     by (cases "x = y", auto simp: subst_def)
 qed
 with Sub show ?thesis by simp 
qed

lemma substRan[simp] : "ran {x\<leftarrow>t} = (if (t = Var x) then {} else {t})"
  by (cases "t = Var x", auto simp: dom.simps ran.simps subst_def)

lemma substIdent[simp] : "s \<cdot> {x\<leftarrow>Var x} = s"
proof (induct s)
  case (Var y) thus ?case by (cases "x = y", auto simp: subst_def)
next
  case (Fun f ss) thus ?case 
  proof (induct ss)
    case Nil thus ?case by (simp add: subst_def)
  next
    case Cons thus ?case by (auto simp:subst_def)
  qed
qed


lemma substIdent2 : assumes "x \<notin> vars_term t" shows "t \<cdot> {x \<leftarrow> u} = t"
using assms
proof -
  have "x \<notin> vars_term t \<Longrightarrow> t \<cdot> Sub (\<lambda>y. if x = y then u else Var y) = t" (is "_ \<Longrightarrow> t \<cdot> ?S = t")
  proof (induct t)
    case (Var y) thus ?case by (cases "x = y", auto)
  next
    case (Fun f ss)
    hence "\<forall> s \<in> set ss. s \<cdot> ?S = s" by auto
    hence mid: "map (\<lambda> v. v \<cdot> ?S) ss = ss"
    proof (induct ss)
      case Nil thus ?case by simp
    next
      case (Cons s sss)
      hence sid: "s \<cdot> ?S = s" by (simp add: mem_iff)
      from Cons have xsss: "\<forall> s \<in> set sss. s \<cdot> ?S = s" by auto
      with Cons and sid show ?case by auto
    qed
    show ?case by (simp add: mid)
  qed
  with `x \<notin> vars_term t` show ?thesis by (unfold subst_def, simp)
qed


lemma substVars : "vars_term (t \<cdot> \<sigma>) \<subseteq> vars_term t - dom \<sigma> \<union> \<Union> vars_term`ran \<sigma> " 
proof (cases \<sigma>)
  case (Sub sigma)
  have "vars_term (t \<cdot> (Sub sigma)) \<subseteq> vars_term t - dom (Sub sigma) \<union> \<Union>vars_term ` Term.ran (Sub sigma)"
  proof (induct t)
    case (Fun f ss) thus ?case by auto
  next
    case (Var y) show ?case 
    proof
      fix x
      assume ass: "x \<in> vars_term(Var y \<cdot> Sub sigma)"
      show "x \<in> vars_term(Var y) - dom (Sub sigma) \<union> \<Union>vars_term ` ran (Sub sigma)" 
      proof (cases "x=y")
	case True with ass show ?thesis by (cases "x \<in> dom (Sub sigma)", auto simp: ran.simps dom.simps)
      next
	case False
	with ass have "sigma y \<noteq> Var y" by (cases "sigma y",auto)
	with ass  show ?thesis by (auto simp: ran.simps dom.simps)
      qed
    qed
  qed
  with Sub show ?thesis by simp
qed

lemma substVars2 : assumes elem: "x \<in> vars_term(t \<cdot> \<sigma>)" shows "x \<in> vars_term t - dom \<sigma> \<union> \<Union> vars_term`ran \<sigma> "
proof -
  have "vars_term(t \<cdot> \<sigma>) \<subseteq> vars_term t - dom \<sigma> \<union> \<Union> vars_term`ran \<sigma> " by (rule substVars)
  with elem show ?thesis by auto
qed
    
lemma substVanish: assumes xt: "x \<notin> vars_term t" shows "x \<notin> vars_term(u \<cdot> {x\<leftarrow>t})"
proof 
  assume xv: "x \<in> vars_term(u \<cdot> {x\<leftarrow>t})"
  show False
  proof (cases "t = Var x")
    case True with xt show False by simp
  next
    case False
    from xv have "x \<in> vars_term u - dom {x\<leftarrow>t} \<union> \<Union> vars_term`ran {x\<leftarrow>t}" by (rule substVars2)
    with False and xt show False by simp
  qed
qed  

lemma substInvert1 : assumes yt: "y \<notin> vars_term t" and ysxt: "y \<in> vars_term(s \<cdot> {x\<leftarrow>t})" shows "y \<in> vars_term s"
using assms 
proof (cases "t = Var x", auto) 
  assume tx: "t \<noteq> Var x" 
  from ysxt show "y \<in> vars_term s"
  proof (induct s)
    case (Var z) with tx and yt show ?case by (cases "z = x", auto simp: subst_def)
  next
    case (Fun f ss) with yt show ?case by (auto simp: subst_def)
  qed
qed

lemma substInvert2 : assumes "y \<notin> vars_term t" and "x \<noteq> y" and "s \<cdot> {x\<leftarrow>t} \<noteq> Var y" shows
  "s \<noteq> Var y" 
using assms
by (cases s, auto simp: subst_def)

lemma substInvert3 : assumes "y \<notin> vars_term t" and "s \<cdot> {x\<leftarrow>t} = Var y" shows "s = Var y" 
using assms
proof (cases s, auto simp: subst_def)
  fix z
  show "\<lbrakk>(if x = z then t else Var z) = Var y; y \<notin> vars_term t\<rbrakk> \<Longrightarrow> z = y"  by (cases "x=z", auto)
qed

lemma substFun[simp] : "(Fun f ss \<cdot> \<sigma> = Fun g ts \<cdot> \<sigma>) = (f = g \<and> length ss = length ts \<and> \<sigma> \<in> unifiers (set (zip ss ts)))" 
using assms
proof (cases \<sigma>)
  case (Sub sigma)
  hence "(Fun f ss \<cdot> (Sub sigma) = Fun g ts \<cdot> (Sub sigma)) = (f = g \<and> length ss = length ts \<and> (Sub sigma) \<in> unifiers (set (zip ss ts)))" 
  proof (induct ss arbitrary: ts)
    case Nil thus ?case by simp
  next
    case (Cons s sss)
    thus ?case by (cases ts, auto simp: unifiersInsert)
  qed
  with Sub show ?thesis by simp
qed


(* here the functions for computing mgu's start *)

(* x occurs somewhere within a list of term-pairs *)
definition varsL :: "(('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> 'v \<Rightarrow> bool"
 where "varsL list x == foldr (\<lambda> pair b. x mem vars_term_list (fst pair) \<or> x mem vars_term_list (snd pair) \<or> b) list False"

(* substitution {x \<leftarrow> t} applied on a term  *)
fun simpleSubst :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) term"
where "simpleSubst x t (Var y) = (if x = y then t else (Var y))"
   |  "simpleSubst x t (Fun f ss) = Fun f (map (simpleSubst x t) ss)"


(* substitution {x \<leftarrow> t} applied on term-pair list *)
definition simpleSubstList :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term) list \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term) list"
where "simpleSubstList x u list = map (\<lambda> p. (simpleSubst x u (fst p), simpleSubst x u (snd p))) list"


definition decompose :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option"
  where "decompose s t = (case (s,t) of (Fun f ss, Fun g ts) \<Rightarrow> if f = g then decomp ss ts else None | _ \<Rightarrow> None)"

(* unification of list of pairs of terms *)
function unify :: "(('f, 'v) term \<times> ('f, 'v) term) list \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term) list \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term) list option"
  where "unify [] seen = Some seen"
      | "unify ((Fun f ss, Fun g ts) # unifs) seen = (case decompose (Fun f ss) (Fun g ts) of None \<Rightarrow> None | Some pairs \<Rightarrow> unify (pairs @ unifs) seen)"
      | "unify ((Fun f ss, Var x) # unifs) seen = unify ((Var x, Fun f ss) # unifs) seen"
      | "unify ((Var x , t) # unifs) seen = (if t = Var x then unify unifs seen
             else if x mem vars_term_list t then None 
             else if varsL (unifs @ seen) x then unify ((simpleSubstList x t (unifs @ seen)) @ [(Var x, t)]) []
             else unify unifs ((Var x, t) # seen))"
by (pat_completeness, auto)

(* extracting a substitution from a list of pairs *)
fun subst_of :: "(('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> ('f,'v) sub"
  where "subst_of [] = \<iota>"
     |  "subst_of ((Var x,t) # ts) = Sub (\<lambda> y. if (x = y) then t else
                      (case subst_of ts of Sub s \<Rightarrow> s y))"


(* computing the mgu of two terms *)
fun mgu :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) sub option"
  where "mgu s t = (case unify [(s,t)] [] of   
                      None \<Rightarrow> None
                    | Some res \<Rightarrow> Some (subst_of res))"

(* the mgu of two terms (to be used if it is guaranteed that the terms are unifiable) *)
definition mgu_of where "mgu_of s t == (case mgu s t of None \<Rightarrow> \<iota> | Some \<delta> \<Rightarrow> \<delta>)" 


(* properties which will be proven later *)
lemma mguComplete : "mgu s t = None \<Longrightarrow> unifiers {(s,t)} = {}"
oops
lemma mguSound : "mgu s t = Some \<delta> \<Longrightarrow> isSMgu \<delta> {(s,t)}"
oops
lemma mgu_of[simp] :  "s \<cdot> \<sigma> = t \<cdot> \<sigma> \<Longrightarrow> s \<cdot> mgu_of s t = t \<cdot> mgu_of s t \<and> \<sigma> = mgu_of s t \<circ>s \<sigma>" 
oops



(* and their properties *)

lemma simpleSubstSound[simp] : "simpleSubst x t = (\<lambda> u. u \<cdot> {x \<leftarrow> t})"
proof (rule ext)
  fix u
  show "simpleSubst x t u = u \<cdot> {x\<leftarrow>t}"
  proof (induct u)
    case (Var y)
    show ?case by (cases "x = y", auto simp:subst_def)
  next
    case (Fun f us) thus ?case by (auto simp:subst_def)
  qed
qed



lemma simpleSubstListSound : "simpleSubstList x t pairs = map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x \<leftarrow> t})) pairs"
  by (unfold simpleSubstList_def, simp)

lemma simpleSubstListSet[simp] : "set (simpleSubstList x t pairs) = (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` (set pairs)"
  by (simp add: simpleSubstListSound)

definition varsPairs where "varsPairs zs == (\<Union> p \<in> zs . vars_term(fst p) \<union> vars_term(snd p))"

lemma varsLSound[simp] : "varsL l x = (x \<in> varsPairs (set l))"
proof -
  have "foldr (\<lambda> pair b. x mem vars_term_list (fst pair) \<or> x mem vars_term_list (snd pair) \<or> b) l False = (x \<in> varsPairs (set l))"
  proof (induct l)
    case Nil thus ?case by (simp add: varsPairs_def)
  next
    case (Cons pair ll)
    thus ?case 
    proof (cases pair)
      case (Pair s t)
      with Cons vars_term_list_sound show ?thesis 
	by (cases "x mem vars_term_list s \<or> x mem vars_term_list t", auto simp: mem_iff varsPairs_def)
    qed
  qed
  thus ?thesis by (simp add: varsL_def)
qed


lemma decomposeZip : "decompose (Fun f ss) (Fun g ts) = Some pairs \<Longrightarrow> pairs = zip ss ts" 
  by (cases "f=g", auto simp: decompose_def decompZip) 
lemma decomposeFg : "decompose (Fun f ss) (Fun g ts) = Some pairs \<Longrightarrow> f = g" 
  by (cases "f=g", auto simp: decompose_def) 
lemma decomposeLength : "decompose (Fun f ss) (Fun g ts) = Some pairs \<Longrightarrow> length ss = length ts"
  by (cases "f=g", auto simp: decompose_def decompLength) 
lemma decomposeNone : "decompose (Fun f ss) (Fun g ts) = None \<Longrightarrow> f \<noteq> g \<or> length ss \<noteq> length ts"
  by (cases "f=g", auto simp: decompose_def decompNone)

lemma decomposeSoundI[simp] : assumes dec: "decompose (Fun f ss) (Fun g ts) = None" shows "unifiers (insert((Fun f ss, Fun g ts)) pairs) = {}"
proof - 
  from dec have "f \<noteq> g \<or> length ss \<noteq> length ts" by (rule decomposeNone) 
  thus ?thesis by (auto simp:unifiersInsert List.map_eq_imp_length_eq[where ?xs = "ss" and ?ys = "ts"])
qed

lemma decomposeSoundII[simp] : assumes dec: "decompose (Fun f ss) (Fun g ts) = Some pairs" 
  shows "unifiers (insert ((Fun f ss, Fun g ts)) rest) = unifiers (set pairs \<union> rest)"
using assms
proof - 
  from dec have fg: "f = g" by (rule decomposeFg) 
  from dec have len: "length ss = length ts" by (rule decomposeLength) 
  from dec have zip: "pairs = zip ss ts" by (rule decomposeZip)
  from len have "{\<sigma>. map (\<lambda>t. t \<cdot> \<sigma>) ss = map (\<lambda>t. t \<cdot> \<sigma>) ts} = unifiers (set (zip ss ts))" 
  proof (induct ss arbitrary: ts)
    case Nil thus ?case by (cases ts, auto simp: unifiers_def)
  next
    case (Cons s ss) note oCons = this
    thus ?case 
    proof (cases ts)
      case Nil with Cons have False by simp thus ?thesis ..
    next
      case (Cons t ts) 
      with oCons show ?thesis by (auto simp: unifiers_def)
    qed
  qed
  with fg zip show ?thesis by (auto simp: unifiersInsert) 
qed



(* towards termination of unify *)

primrec nvars :: "('f,'v) term \<Rightarrow> 'v set"
  where "nvars (Var _) = {}"
     |  "nvars (Fun f ss) = vars_term(Fun f ss)"

definition doubleVars :: "(('f,'v) term \<times> ('f,'v) term) set \<Rightarrow> 'v set"
  where "doubleVars prob == { x | p q x. p \<in> prob \<and> q \<in> prob \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q}"

lemma doubleVarsMono[simp] : assumes "s \<subseteq> t" shows "doubleVars s \<subseteq> doubleVars t"
using assms
proof (unfold doubleVars_def)
  let ?S = "\<lambda> p q x u. u = x \<and> p \<in> s \<and> q \<in> s \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q" 
  let ?L = "\<lambda> p q x u. u = x \<and> p \<in> t \<and> q \<in> t \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q"
  show "{u . \<exists> p q x. ?S p q x u} \<subseteq> {u . \<exists> p q x. ?L p q x u}"
  proof 
    fix y
    assume "y \<in> {u . \<exists> p q x. ?S p q x u}"
    hence "\<exists> p q x. ?S p q x y" by auto
    from this obtain p q x where "?S p q x y" by blast
    with  assms have "?L p q x y" by auto
    hence "\<exists> p q x. ?L p q x y" by blast
    thus "y \<in> {u. \<exists> p q x. ?L p q x u}" by simp
  qed
qed

lemma doubleVarsMono2[simp] : assumes sub: "s \<subseteq> t" and elem: "x \<in> doubleVars s" shows "x \<in> doubleVars t"
proof -
  from sub have "doubleVars s \<subseteq> doubleVars t" by simp
  with elem show "x \<in> doubleVars t" by auto
qed
  

lemma doubleVarsFun[simp] : "doubleVars ({(Fun f ss,t)} \<union> prob) = doubleVars prob" 
proof 
  have "prob \<subseteq> {(Fun f ss,t)} \<union> prob" by auto
  thus "doubleVars prob \<subseteq> doubleVars ({(Fun f ss,t)} \<union> prob)" by simp
next
  show "doubleVars ({(Fun f ss,t)} \<union> prob) \<subseteq> doubleVars prob"
  proof (unfold doubleVars_def)
    let ?S = "\<lambda> p q x u. u = x \<and> p \<in> prob \<and> q \<in> prob \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q" 
    let ?L = "\<lambda> p q x u. u = x \<and> p \<in> {(Fun f ss, t)} \<union> prob \<and> q \<in> {(Fun f ss, t)} \<union> prob \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q"
    show "{u . \<exists> p q x. ?L p q x u} \<subseteq> {u . \<exists> p q x. ?S p q x u}"
    proof
      fix y
      assume "y \<in> {u . \<exists> p q x. ?L p q x u}"
      hence "\<exists> p q x. ?L p q x y" by auto
      from this obtain p q x where "?L p q x y" by blast
      hence "?S p q x y" by auto
      hence "\<exists> p q x. ?S p q x y" by blast
      thus "y \<in> {u. \<exists> p q x. ?S p q x u}" by simp
    qed
  qed
qed
  
definition unsolvedVars :: "(('f,'v) term \<times> ('f,'v) term) set \<Rightarrow> 'v set"
where "unsolvedVars prob == (\<Union> (\<lambda> p. nvars (fst p)) ` prob \<union> \<Union> (\<lambda> p. vars_term(snd p)) ` prob  \<union> doubleVars prob)"

lemma unsolvedVarsMono[simp] : assumes sub: "s \<subseteq> t" shows "unsolvedVars s \<subseteq> unsolvedVars t"
using assms 
by (unfold unsolvedVars_def, auto)

lemma nvarsVars : "nvars t \<subseteq> vars_term t" by (cases t) auto

lemma nvarsVars2[simp] : "x \<in> nvars t \<Longrightarrow> x \<in> vars_term t" by (cases t) auto

lemma doubleVarsVars : "doubleVars s \<subseteq> (\<Union> p \<in> s. vars_term(fst p) \<union> vars_term(snd p))"
proof (unfold doubleVars_def, rule subsetI)
  fix x
  assume ass: "x \<in> {u. \<exists>p q x. u = x \<and> p \<in> s \<and> q \<in> s \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q}"
  from this have "\<exists>p q . p \<in> s \<and> q \<in> s \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q" by blast
  from this obtain p where p: "p \<in> s" and fst: "fst p = Var x" by blast
  hence "x \<in> vars_term(fst p)" by simp
  with p fst show "x \<in> (\<Union>p\<in>s. vars_term(fst p) \<union> vars_term(snd p))" by blast
qed

lemma unsolvedVarsVars : "unsolvedVars s \<subseteq> (\<Union> p \<in> s. vars_term(fst p) \<union> vars_term(snd p))" (is "_ \<subseteq> ?S")
proof -
  have "doubleVars s \<subseteq> ?S" by (rule doubleVarsVars)
  thus ?thesis by (auto simp: unsolvedVars_def)
qed

lemma finiteVars : "finite (\<Union> p \<in> set s. vars_term(fst p) \<union> vars_term(snd p))"
proof (rule finite_UN_I)
  show "finite (set s)" by simp
next
  fix p
  show "finite (vars_term(fst p) \<union> vars_term(snd p))"
  proof (rule finite_UnI)
    show "finite (vars_term(fst p))" by simp
    show "finite (vars_term(snd p))" by simp
  qed
qed

lemma unsolvedVarsFinite : "finite (unsolvedVars (set list))" 
proof (rule finite_subset)
  show "unsolvedVars (set list) \<subseteq> (\<Union> p \<in> set list. vars_term(fst p) \<union> vars_term(snd p))" by (rule unsolvedVarsVars)
next
  show "finite (\<Union>p\<in>set list. vars_term(fst p) \<union> vars_term(snd p))" by (rule finiteVars)
qed

definition m1 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where "m1 = (\<lambda> (l1,l2). card (unsolvedVars (set(l1) \<union> set(l2))))"

definition m1a :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where "m1a = (\<lambda> (l1,l2). length (filter (\<lambda> p. is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars (set l1 \<union> set l2)) (l1 @ l2)))"

definition m2 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where "m2 = (\<lambda> (l1,l2). (listsum (map (\<lambda> (s,t). |s| + |t| ) l1) + (listsum (map (\<lambda> (s,t). |s| + |t| )  l2))))"

definition m3 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where "m3 = (\<lambda> (l1,l2). length (filter (\<lambda> (s,t). \<not> is_var s \<and> is_var t) (l1 @ l2)))"

definition m4 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
  where "m4 = (\<lambda> (l1, l2) . length l1)"

(* termination can be shown by lex. combination of (m1,m1a,m2,m3,m4) *)
  

lemma unsolvedVarsFinite2: "finite (unsolvedVars (set l \<union> set r))"
proof -
  have "finite (unsolvedVars (set (l @ r)))" by (rule unsolvedVarsFinite)
  thus ?thesis by simp
qed



lemma m1subset :  "unsolvedVars (set (r1) \<union> set (r2)) \<subseteq> unsolvedVars (set (l1) \<union> set (l2)) \<Longrightarrow> m1 (r1,r2) \<le> m1 (l1,l2)" 
proof -
  assume sub: "unsolvedVars (set (r1) \<union> set (r2)) \<subseteq>  unsolvedVars (set (l1) \<union> set (l2))" (is "?S \<subseteq> ?L")
  have "card ?S \<le> card ?L"
  proof (rule card_mono)
    show "finite (unsolvedVars (set l1 \<union> set l2))" by (rule unsolvedVarsFinite2) 
    from sub show "?S \<subseteq> ?L" .
  qed
  thus ?thesis by (unfold m1_def, auto)
qed


lemma m1psubset :  "unsolvedVars (set (r1) \<union> set (r2)) \<subset> unsolvedVars (set (l1) \<union> set (l2)) \<Longrightarrow> m1 (r1,r2) < m1 (l1,l2)" 
proof -
  assume sub: "unsolvedVars (set (r1) \<union> set (r2)) \<subset> unsolvedVars (set (l1) \<union> set (l2))" (is "?S \<subset> ?L")
  have "card ?S < card ?L"
  proof (rule psubset_card_mono)
    show "finite (unsolvedVars (set l1 \<union> set l2))" by (rule unsolvedVarsFinite2) 
    from sub show "?S \<subset> ?L" .
  qed
  thus ?thesis by (unfold m1_def, auto)
qed
  
  
lemma doubleVarsSubst : assumes yt: "y \<notin> vars_term t" 
  and rs: "y \<in> doubleVars ((\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` pairs)" (is "_ \<in> doubleVars ?XTP")
  shows "y \<in> doubleVars pairs"
using rs
proof (unfold doubleVars_def)
  let ?S = "\<lambda> p q z u. u = z \<and>
                       p \<in> ?XTP \<and> q \<in> ?XTP \<and> fst p = Var z \<and> fst q = Var z \<and> p \<noteq> q"
  let ?O = "\<lambda> p q z u. u = z \<and>
                       p \<in> pairs \<and> q \<in> pairs \<and> fst p = Var z \<and> fst q = Var z \<and> p \<noteq> q"
  assume "y \<in> {u. \<exists> p q z. ?S p q z u}" 
  from this have  "\<exists> p q z. ?S p q z y" by blast
  from this obtain p q where cond: "?S p q y y" by blast
  hence "fst p = Var y" and "fst q = Var y" by auto
  from cond have "p \<in> ?XTP" by simp
  hence "\<exists> opair. opair \<in> pairs \<and> p = (fst opair \<cdot> {x\<leftarrow>t}, snd opair \<cdot> {x\<leftarrow>t})" by (auto, blast)
  from this obtain opair where piece1: "opair \<in> pairs" and pp: "p = (fst opair \<cdot> {x\<leftarrow>t}, snd opair \<cdot> {x\<leftarrow>t})" by blast 
  from pp and cond have "fst opair \<cdot> {x\<leftarrow>t} = Var y" by simp
  with yt have piece2: "(fst opair) = Var y" by (rule substInvert3)
  from cond have "q \<in> ?XTP" by simp
  hence "\<exists> opair2. opair2 \<in> pairs \<and> q = (fst opair2 \<cdot> {x\<leftarrow>t}, snd opair2 \<cdot> {x\<leftarrow>t})" by (auto, blast)
  from this obtain opair2 where piece3: "opair2 \<in> pairs" and qq: "q = (fst opair2 \<cdot> {x\<leftarrow>t}, snd opair2 \<cdot> {x\<leftarrow>t})" by blast
  from qq and cond have "fst opair2 \<cdot> {x\<leftarrow>t} = Var y" by simp
  with yt have piece4: "(fst opair2) = Var y" by (rule substInvert3)
  from piece1 piece2 piece3 piece4 and cond and pp and qq have "?O opair opair2 y y" by auto
  hence "\<exists> p q z. ?O p q z y" by blast
  thus "y \<in> {u. \<exists> p q z. ?O p q z u}" by blast
qed

lemma fstZip : "x \<in> set (zip as bs) \<Longrightarrow> fst x \<in> set as"
proof (induct as arbitrary: bs)
  case Nil thus ?case by simp
next
  case (Cons c cs) thus ?case by (cases bs, auto)
qed

lemma sndZip : "p \<in> set (zip as bs) \<Longrightarrow> snd p \<in> set bs"
proof (induct bs arbitrary: as)
  case Nil thus ?case by simp
next
  case Cons thus ?case by (cases as, auto)
qed

lemma varsPairsExtract : "x \<in> varsPairs pairs \<Longrightarrow> \<exists> p \<in> pairs. x \<in> vars_term (fst p) \<or> x \<in> vars_term (snd p)"
  by (unfold varsPairs_def, auto) 

lemma varsPairsUnion[simp] : "varsPairs (ps \<union> qs) = varsPairs ps \<union> varsPairs qs" 
  by (unfold varsPairs_def, auto)

lemma varsPairsInsert[simp] : "varsPairs (insert pair ps) = vars_term(fst pair) \<union> vars_term(snd pair) \<union> varsPairs ps"
  by (unfold varsPairs_def, auto)

lemma varsPairsEmpty[simp] : "varsPairs {} = {}"
  by (unfold varsPairs_def, auto)

lemma varsPairsFun[simp] : "x \<in> varsPairs (set (zip ss ts)) \<Longrightarrow> x \<in> vars_term(Fun f ss) \<or> x \<in> vars_term(Fun g ts)" 
proof (unfold varsPairs_def, induct ss arbitrary: ts)
  case Nil thus ?case by simp
next
  case (Cons s ss) note oCons = this
  thus ?case
  proof (cases ts)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons t tts)
    with oCons show ?thesis by auto
  qed
qed

termination unify
proof -
  let ?R = "measures [m1,m1a,m2,m3,m4]"

  have m1step: "!! l1 l2 r1 r2.
    \<lbrakk>unsolvedVars (set r1 \<union> set r2) \<subseteq> unsolvedVars (set l1 \<union> set l2); ((r1,r2),l1,l2) \<in> measures [m2,m3,m4];
    unsolvedVars (set r1 \<union> set r2) = unsolvedVars (set l1 \<union> set l2) \<Longrightarrow> m1a (r1,r2) \<le> m1a (l1,l2)\<rbrakk> 
    \<Longrightarrow> ((r1,r2),l1,l2) \<in> ?R" 
  proof -
    fix l1 :: "(('f,'v) term \<times> ('f,'v) term) list" and 
      l2 :: "(('f,'v) term \<times> ('f,'v) term) list" and 
      r1 :: "(('f,'v) term \<times> ('f,'v) term) list" and 
      r2 :: "(('f,'v) term \<times> ('f,'v) term) list" 
    assume uns: "unsolvedVars (set r1 \<union> set r2) \<subseteq> unsolvedVars (set l1 \<union> set l2)"
    assume mea: "((r1,r2),l1,l2) \<in> measures [m2,m3,m4]"
    assume m1get: "unsolvedVars (set r1 \<union> set r2) = unsolvedVars (set l1 \<union> set l2) \<Longrightarrow> m1a (r1,r2) \<le> m1a (l1,l2)"
    show "((r1,r2),l1,l2) \<in> ?R"
    proof (cases "unsolvedVars (set r1 \<union> set r2) \<subset> unsolvedVars (set l1 \<union> set l2)")
      case True
      hence "m1 (r1,r2) < m1 (l1,l2)" by (rule m1psubset)
      thus ?thesis by (rule measures_less)
    next
      case False 
      from uns have "m1 (r1,r2) \<le> m1 (l1,l2)" by (rule m1subset)
      thus ?thesis 
      proof (rule measures_lesseq) 
	from uns and m1get and False have "m1a (r1,r2) \<le> m1a (l1,l2)" by auto 
	from this and mea show "((r1,r2),l1,l2) \<in> measures [m1a,m2,m3,m4]"
	  by (rule measures_lesseq)
      qed
    qed
  qed

  show ?thesis 
  proof
  
    show "wf ?R" by simp
  
      (* case 5 *)
    fix x t unifs seen 
    show "((unifs, (Var x, t) # seen), (Var x, t) # unifs, seen) \<in> ?R"
    proof (rule m1step)
      show "unsolvedVars (set unifs \<union> set ((Var x, t) # seen)) \<subseteq> unsolvedVars (set ((Var x, t) # unifs) \<union> set seen)" by simp
    next
      show "((unifs, (Var x, t) # seen), (Var x, t) # unifs, seen) \<in> measures [m2, m3, m4]"
      proof (rule measures_lesseq)
	show "m2 (unifs, (Var x, t) # seen) \<le> m2 ((Var x, t) # unifs, seen)"
	  by (unfold m2_def, simp)
      next
	show "((unifs, (Var x, t) # seen), (Var x, t) # unifs, seen) \<in> measures [m3, m4]"
	proof (rule measures_lesseq)
	  show "m3 (unifs, (Var x, t) # seen) \<le> m3 ((Var x, t) # unifs, seen)"
	    by (unfold m3_def, simp)
	next
	  show "((unifs, (Var x, t) # seen), (Var x, t) # unifs, seen) \<in> measures [m4]"
            by (rule measures_less, unfold m4_def, simp)
	qed
      qed
    next
      show "m1a (unifs, (Var x, t) # seen) \<le> m1a ((Var x, t) # unifs, seen)" 
	by (unfold m1a_def, auto)
    qed

  next
    
      (* case 3 *)
    fix x t unifs seen
    show "((unifs, seen), (Var x, t) # unifs, seen) \<in> measures [m1,m1a,m2,m3,m4]"
    proof (rule m1step) 
      have "set unifs \<union> set seen \<subseteq> set ((Var x, t) # unifs) \<union> set seen" by auto
      thus "unsolvedVars (set unifs \<union> set seen) \<subseteq> unsolvedVars (set ((Var x, t) # unifs) \<union> set seen)" by auto
    next
      show "((unifs, seen), (Var x, t) # unifs, seen) \<in> measures [m2,m3,m4]"
	by (rule measures_less, unfold m2_def, simp)
    next
      assume "unsolvedVars (set unifs \<union> set seen) = unsolvedVars (set ((Var x, t) # unifs) \<union> set seen)"
      thus "m1a (unifs, seen) \<le> m1a ((Var x, t) # unifs, seen)" by (unfold m1a_def, auto)
    qed

  next

      (* case 2 *)
    fix f ss x unifs seen 
    show "(((Var x, Fun f ss) # unifs, seen), (Fun f ss, Var x) # unifs, seen) \<in> measures [m1, m1a, m2, m3, m4]"
    proof (rule m1step)
      show "(((Var x, Fun f ss) # unifs, seen), (Fun f ss, Var x) # unifs, seen) \<in> measures [m2, m3, m4]"
      proof (rule measures_lesseq)
	show "(((Var x, Fun f ss) # unifs, seen), (Fun f ss, Var x) # unifs, seen) \<in> measures [m3, m4]"
          by (rule measures_less, unfold m3_def, simp)
      next
	show "m2 ((Var x, Fun f ss) # unifs, seen) \<le> m2 ((Fun f ss, Var x) # unifs, seen)"
	  by (unfold m2_def, simp)
      qed
    next
      show "unsolvedVars (set ((Var x, Fun f ss) # unifs) \<union> set seen) \<subseteq> unsolvedVars (set ((Fun f ss, Var x) # unifs) \<union> set seen)"
	by (auto simp: unsolvedVars_def doubleVars_def)
    next
      assume "unsolvedVars (set ((Var x, Fun f ss) # unifs) \<union> set seen) = unsolvedVars (set ((Fun f ss, Var x) # unifs) \<union> set seen)"
      thus "m1a ((Var x, Fun f ss) # unifs, seen) \<le> m1a ((Fun f ss, Var x) # unifs, seen)"
      proof (unfold m1a_def, simp)
	show "x \<in> unsolvedVars (insert (Fun f ss, Var x) (set unifs \<union> set seen))" 
	  by (unfold unsolvedVars_def, simp)
      qed
    qed
      
  next

      (* case 1 *)
    
    fix f ss g ts unifs seen pairs
    have main: "((zip ss ts @ unifs, seen), (Fun f ss, Fun g ts) # unifs, seen) \<in> ?R"
    proof (rule m1step)
      show "((zip ss ts @ unifs, seen), (Fun f ss, Fun g ts) # unifs, seen) \<in> measures [m2, m3, m4]"
      proof (rule measures_less, unfold m2_def, simp, induct ss arbitrary: ts, simp)
	case (Cons b sss) 
	assume A: "\<And>ts. (\<Sum>(s, t)\<leftarrow>zip sss ts. |s| + |t| ) < Suc (Suc (listsum (map sz sss) + listsum (map sz ts)))"
	thus ?case 
	proof (cases ts, simp)
	  case (Cons c ttt)
	  from `(\<Sum>(s, t)\<leftarrow>zip sss ttt. |s| + |t| ) < Suc (Suc (listsum (map sz sss) + listsum (map sz ttt)))`
            and Cons
	  show ?thesis by simp
	qed
      qed
    next
      assume eq: "unsolvedVars (set (zip ss ts @ unifs) \<union> set seen) = unsolvedVars (set ((Fun f ss, Fun g ts) # unifs) \<union> set seen)"
      let ?ZIP = "zip ss ts"
      let ?ZIPUS = "set (?ZIP @ unifs) \<union> set seen"
      let ?C = "\<lambda> p.  is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars (set (zip ss ts @ unifs) \<union> set seen)"
      have "set [p\<leftarrow>zip ss ts. is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars (set (zip ss ts @ unifs) \<union> set seen)] = {}" (is "set ?PS = {}")
      proof (cases ?PS)
	case Nil
	thus ?thesis by simp
      next
	case (Cons p ps)
	hence "p \<in> set ?PS" by auto
	hence "p \<in> {p. p \<in> set ?ZIP \<and> is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars (set (zip ss ts @ unifs) \<union> set seen)}" by simp
	from this obtain p where pzip: "p \<in> set ?ZIP" and pcond: "is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars (set (zip ss ts @ unifs) \<union> set seen)" by blast
	from pcond obtain x where xp: "fst p = Var x" by (cases "fst p", auto)
	with pcond have xnus: "x \<notin> unsolvedVars (set (zip ss ts @ unifs) \<union> set seen)" by simp
	have "x \<in> nvars (Fun f ss)"
	proof -
	  from xp have xFirst: "x \<in> vars_term(fst p)" by simp
	  from pzip have "fst p \<in> set ss" by (rule fstZip)
	  with xFirst show "x \<in> nvars (Fun f ss)" by auto
	qed
	hence "x \<in> unsolvedVars (set ((Fun f ss, Fun g ts) # unifs) \<union> set seen)"
	  by (unfold unsolvedVars_def, auto)
	with xnus and eq have False by simp thus ?thesis ..
      qed
      
      hence "length [p \<leftarrow> zip ss ts. ?C p] = 0" by (cases "[p \<leftarrow> zip ss ts. ?C p]", auto)
      with eq show "m1a (zip ss ts @ unifs, seen) \<le> m1a ((Fun f ss, Fun g ts) # unifs, seen)" 
	by (unfold m1a_def, auto) 	      
    next
      let ?Z = "set (zip ss ts)"
      let ?US = "set (unifs) \<union> set seen"
      let ?F = "{(Fun f ss, Fun g ts)}"
      let ?ZUS = "?Z \<union> ?US"
      let ?FUS = "?F \<union> ?US"
      have main2: "unsolvedVars (?ZUS) \<subseteq> unsolvedVars (?FUS)"
      proof (auto simp: unsolvedVars_def)
	fix x a b 
	assume elem: "(a,b) \<in> ?Z" 
	assume "x \<in> nvars a" hence elem2: "x \<in> vars_term a" by simp 
	assume ss: "\<forall> c \<in> set ss. x \<notin> vars_term c"
	from elem have "fst (a,b) \<in> set ss" by (rule fstZip)
	with ss and elem2 have False by auto thus "\<exists>xa\<in>set unifs. x \<in> nvars (fst xa)" ..
      next
	fix x a b 
	assume elem: "(a,b) \<in> ?Z" and elem2: "x \<in> vars_term b" and ts: "\<forall> c \<in> set ts. x \<notin> vars_term c"
	from elem have "snd (a,b) \<in> set ts" by (rule sndZip)
	with elem2 and ts have False by simp thus "\<exists>xa\<in>set unifs. x \<in> nvars(fst xa)" ..
      next
	fix y
	assume yD: "y \<in> doubleVars ?ZUS" and 
	  ynD: "y \<notin> doubleVars (insert (Fun f ss, Fun g ts) (set unifs \<union> set seen))" and
          "\<forall>ya\<in>set ts. y \<notin> vars_term ya" and
	  "\<forall>ya\<in>set ss. y \<notin> vars_term ya" 
	let ?S = "\<lambda> p q x u. u = x \<and> p \<in> ?ZUS \<and> q \<in> ?ZUS \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q" 
	let ?L = "\<lambda> p q x u. u = x \<and> p \<in> ?FUS \<and> q \<in> ?FUS \<and> fst p = Var x \<and> fst q = Var x \<and> p \<noteq> q"
	from yD have "\<exists> p q x. ?S p q x y" by (unfold doubleVars_def, auto)
	from this obtain p q x where pqx: "?S p q x y" by blast
	have step: "!! r. \<lbrakk>fst r = Var y; r \<in> ?ZUS\<rbrakk> \<Longrightarrow> r \<in> ?FUS" 
	proof -
	  fix r
	  assume rZUS: "r \<in> ?ZUS" 
	  assume ry: "fst r = Var y"
	  from rZUS show "r \<in> ?FUS"
	  proof 
	    assume "r \<in> ?Z"
	    hence "fst r \<in> set ss" by (rule fstZip)
	    with `\<forall>ya\<in>set ss. y \<notin> vars_term ya` and ry have False by auto thus ?thesis ..
	  next
	    assume "r \<in> ?US" thus ?thesis by simp
	  qed
	qed
	from step and pqx have "?L p q x y" by auto
	hence "\<exists> p q x. ?L p q x y" by blast
	with ynD have False by (unfold doubleVars_def, auto)
	thus "\<exists>x \<in>set unifs. y \<in> nvars (fst x)" ..
      qed
      have "set (zip ss ts) \<union> (set unifs \<union> set seen) = set (zip ss ts @ unifs) \<union> set seen" and
	"{(Fun f ss, Fun g ts)} \<union> (set unifs \<union> set seen) = set ((Fun f ss, Fun g ts) # unifs) \<union> set seen" by auto
      with main2
      show "unsolvedVars (set (zip ss ts @ unifs) \<union> set seen) \<subseteq> unsolvedVars (set ((Fun f ss, Fun g ts) # unifs) \<union> set seen)" by auto
    qed
      
    assume "decompose (Fun f ss) (Fun g ts) = Some pairs"
    hence "pairs = zip ss ts" by (rule decomposeZip)
    with main show "((pairs @ unifs, seen), (Fun f ss, Fun g ts) # unifs, seen) \<in> ?R" by simp
    
  next

      (* case 4 *)
    fix x :: 'v and  t  :: "('f, 'v) term" and unifs :: "(('f,'v) term \<times> ('f,'v) term) list" and seen :: "(('f,'v) term \<times> ('f,'v) term) list"
    assume "varsL (unifs @ seen) x"
    hence "x \<in> varsPairs (set unifs \<union> set seen)" by auto
    hence xUS: "\<exists> p \<in> set unifs \<union> set seen. x \<in> vars_term(fst p) \<or> x \<in> vars_term(snd p)" by (rule varsPairsExtract)
    assume tnx: "t \<noteq> Var x"
    assume "\<not> x mem vars_term_list t" 
    hence xnt: "x \<notin> vars_term t" using vars_term_list_sound by (auto simp: mem_iff)
    let ?US = "set (unifs @ seen)"
    let ?SUS = "(\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` ?US"
    let ?X = "{(Var x, t)}"
    let ?XUS = "?X \<union> ?US" 
    let ?SUSX = "?SUS \<union> ?X"
    let ?P = "\<lambda> (p :: ('f,'v) term \<times> ('f,'v) term) (q :: ('f,'v) term \<times> ('f,'v) term) (s :: (('f,'v) term \<times> ('f,'v) term) set) (z :: 'v). 
      p \<in> s \<and> q \<in> s \<and> fst p = Var z \<and> fst q = Var z \<and> p \<noteq> q" 
    let ?Q = "\<lambda> (s :: (('f,'v) term \<times> ('f,'v) term) set) (z :: 'v). 
      \<exists> p q . p \<in> s \<and> q \<in> s \<and> fst p = Var z \<and> fst q = Var z \<and> p \<noteq> q" 
    
    from xnt have vanish1: "!!s. x \<notin> vars_term(s \<cdot> {x\<leftarrow>t})" by (auto simp:substVanish)
    hence bigVanish1: "!!f s. x \<notin> \<Union> (\<lambda> p. vars_term(f p \<cdot> {x\<leftarrow>t})) ` s" by blast
    have vanish2: "!!s. x \<notin> nvars (s \<cdot> {x\<leftarrow>t})" 
    proof 
      fix s
      assume "x \<in> nvars (s \<cdot> {x\<leftarrow>t})"
      hence "x \<in> vars_term(s \<cdot> {x\<leftarrow>t})" by simp
      with vanish1 show False by blast
    qed
    hence bigVanish2: "!!f s. x \<notin> \<Union> (\<lambda> p. nvars (f p \<cdot> {x\<leftarrow>t})) ` s" by blast
    
    
    have xNotUnsolved: "x \<notin> unsolvedVars ?SUSX"
    proof 
      assume xsus: "x \<in> unsolvedVars ?SUSX"
      from xsus show False
      proof (unfold unsolvedVars_def)
	assume "x \<in> \<Union>(\<lambda>p. nvars (fst p)) ` ?SUSX \<union>  \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUSX  \<union>   doubleVars ?SUSX"
	thus False
	proof
	  assume "x \<in> \<Union>(\<lambda>p. nvars (fst p)) ` ?SUSX \<union> \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUSX"
	  with xnt and vanish1 and vanish2 show False  by (auto)
	next
	  assume "x \<in> doubleVars ?SUSX"
	  hence "\<exists> p q. ?P p q ?SUSX x" by (unfold doubleVars_def, auto)
	  from this obtain p where "\<exists> q. ?P p q ?SUSX x" .. 
	  from this obtain q where "?P p q ?SUSX x" ..
	  hence "\<exists> r. r \<in> ?SUS \<and> fst r = Var x" by blast
	  from this obtain r where rsus: "r \<in> ?SUS" and rx: "fst r = Var x" by blast
	  from rx have "x \<in> vars_term(fst r)" by simp
	  with rsus and vanish1 show False by auto
	qed
      qed
    qed  
    
    
    have subset: "unsolvedVars ?SUSX \<subseteq> unsolvedVars ?XUS"
    proof
      fix y
      assume "y \<in> unsolvedVars ?SUSX"
      with xNotUnsolved have xy: "x \<noteq> y" by auto
      show "y \<in> unsolvedVars ?XUS"
      proof (cases "y \<in> vars_term t")
	case True thus "y \<in> unsolvedVars ?XUS" by (unfold unsolvedVars_def, auto)
      next
	case False
	have "?US \<subseteq> ?XUS" by auto
	from `y \<in> unsolvedVars ?SUSX` have "y \<in> \<Union>(\<lambda>p. nvars (fst p)) ` ?SUSX \<or> y \<in> \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUSX \<or> y \<in> doubleVars ?SUSX"
	  by (unfold unsolvedVars_def, auto)
	thus "y \<in> unsolvedVars ?XUS"
	proof 
	  assume "y \<in> \<Union>(\<lambda>p. nvars (fst p)) ` ?SUSX"
	  with `y \<notin> vars_term t` have "y \<in> \<Union>(\<lambda>p. nvars (fst p)) ` ?SUS " by auto
	  hence "\<exists> p. p \<in> ?SUS \<and> y \<in> nvars (fst p)" by auto
	  from this obtain p where psus: "p \<in> ?SUS" and yn: "y \<in> nvars (fst p)" by blast
	  hence yp: "y \<in> vars_term(fst p)" by simp
	  from psus have "\<exists> q. q \<in> ?US \<and> fst p = fst q \<cdot> {x\<leftarrow>t}" by auto
	  from this obtain q where qus: "q \<in> ?US" and eq: "fst p = fst q \<cdot> {x\<leftarrow>t}" by blast
	  from eq and yp have "y \<in> vars_term(fst q \<cdot> {x\<leftarrow>t})" by simp
	  from `y \<notin> vars_term t` and this have yq: "y \<in> vars_term(fst q)"  by (rule substInvert1)
	  from eq yq xy yn have "y \<in> nvars (fst q)" by (cases "fst q", auto simp: subst_def)
	  with qus show ?thesis by (unfold unsolvedVars_def, auto) 
	next
	  assume "y \<in> \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUSX \<or> y \<in> doubleVars ?SUSX"
	  thus ?thesis
	  proof
	    assume "y \<in> \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUSX"
	    with `y \<notin> vars_term t` have "y \<in> \<Union>(\<lambda>p. vars_term(snd p)) ` ?SUS" by auto
	    hence "\<exists> p. p \<in> ?SUS \<and> y \<in> vars_term(snd p)" by auto
	    from this obtain p where psus: "p \<in> ?SUS" and yp: "y \<in> vars_term(snd p)" by blast
	    from psus obtain q where q: "q \<in> ?US \<and> p = (fst q \<cdot> {x\<leftarrow>t}, snd q \<cdot> {x\<leftarrow>t})" by blast
	    with yp have "y \<in> vars_term(snd q \<cdot> {x\<leftarrow>t})" by simp
	    from `y \<notin> vars_term t` and this have yq: "y \<in> vars_term(snd q)"  by (rule substInvert1)
	    with q show ?thesis by (unfold unsolvedVars_def, auto) 
	  next
	    assume "y \<in> doubleVars ?SUSX"
	    hence "?Q ?SUSX y" by (unfold doubleVars_def, auto)
	    from this obtain p where "\<exists> q. ?P p q ?SUSX y" ..
	    from this obtain q where cond: "?P p q ?SUSX y" ..
	    with xy have ps: "p \<in> ?SUS" and qs: "q \<in> ?SUS" by auto
	    from ps have "\<exists> opair. opair \<in> ?US \<and> p = (fst opair \<cdot> {x\<leftarrow>t}, snd opair \<cdot> {x\<leftarrow>t})" by blast
	    from this obtain opair where opU: "opair \<in> ?US" and pss: "p = (fst opair \<cdot> {x\<leftarrow>t}, snd opair \<cdot> {x\<leftarrow>t})" by blast
	    with cond have "(fst opair) \<cdot> {x\<leftarrow>t} = Var y" by simp
	    with `y \<notin> vars_term t` have pp: "fst opair = Var y" by (rule substInvert3)
	    from qs have "\<exists> opair2. opair2 \<in> ?US \<and> q = (fst opair2 \<cdot> {x\<leftarrow>t}, snd opair2 \<cdot> {x\<leftarrow>t})"  by blast
	    from this obtain opair2 where op2U: "opair2 \<in> ?US" and qss: "q = (fst opair2 \<cdot> {x\<leftarrow>t}, snd opair2 \<cdot> {x\<leftarrow>t})" by blast
	    with cond have "(fst opair2) \<cdot> {x\<leftarrow>t} = Var y" by simp
	    with `y \<notin> vars_term t` have qq: "fst opair2 = Var y" by (rule substInvert3)
	    from cond and pss and qss have "opair \<noteq> opair2" by auto
	    with pp and qq  and opU and op2U have "?P opair opair2 ?XUS y" by auto 
	    hence "y \<in> doubleVars ?XUS" by (unfold doubleVars_def, blast) 
	    thus "y \<in> unsolvedVars ?XUS" by (unfold unsolvedVars_def, auto)
	  qed
	qed
      qed
    qed
    
    show "((simpleSubstList x t (unifs @ seen) @ [(Var x, t)], []), (Var x, t) # unifs, seen) \<in> ?R"
    proof (cases "unsolvedVars ?SUSX \<subset> unsolvedVars ?XUS \<or> (Var x, t) \<notin> ?US \<or> x \<in> unsolvedVars ?XUS")
      case True
      hence psubset: "unsolvedVars ?SUSX \<subset> unsolvedVars ?XUS"
      proof
	assume "unsolvedVars ?SUSX \<subset> unsolvedVars ?XUS" thus "unsolvedVars ?SUSX \<subset> unsolvedVars ?XUS" .
      next
	assume "(Var x, t) \<notin> ?US \<or> x \<in> unsolvedVars ?XUS"
	hence xUnsolved: "x \<in> unsolvedVars ?XUS"
	proof
	  assume "x \<in> unsolvedVars ?XUS" hence "x \<in> unsolvedVars ?XUS" .
	next
	  assume "(Var x,t) \<notin> ?US"
	  show "x \<in> unsolvedVars ?XUS" 
	  proof -
	    from xUS obtain p where pu: "p \<in> set unifs \<union> set seen" and cond: "x \<in> vars_term(fst p) \<or> x \<in> vars_term(snd p)" by auto
	    from cond show "x \<in> unsolvedVars ?XUS"
	    proof
	      assume "x \<in> vars_term(snd p)" with pu show "x \<in> unsolvedVars ?XUS" by (unfold unsolvedVars_def, auto)
	    next
	      assume "x \<in> vars_term(fst p)" 
	      thus "x \<in> unsolvedVars ?XUS"
	      proof (cases "fst p = Var x")
		case False
		with `x \<in> vars_term(fst p)` and pu have "x \<in> nvars (fst p)" by (cases "fst p", auto)
		with pu show "x \<in> unsolvedVars ?XUS" by (unfold unsolvedVars_def, auto)
	      next
		case True
		from `(Var x,t) \<notin> ?US` and pu have "p \<noteq> (Var x, t)" by auto
		with pu and `x \<in> vars_term(fst p)` and True have "?P p (Var x, t) ?XUS x" by auto
		thus "x \<in> unsolvedVars ?XUS" by (unfold unsolvedVars_def doubleVars_def, blast)
	      qed
	    qed
	  qed
	qed
	with subset and xNotUnsolved show "unsolvedVars ?SUSX \<subset> unsolvedVars ?XUS" by auto
      qed
      
      have eq1: "?SUSX = set (simpleSubstList x t (unifs @ seen) @ [(Var x, t)]) \<union> set []" (is "_ = set (?R1) \<union> set (?R2)") by simp
      have eq2: "?XUS = set ((Var x,t) # unifs) \<union> set seen" (is "_ = set(?L1) \<union> set(?L2)") by simp
      from eq1 and eq2 and psubset have "unsolvedVars (set ?R1 \<union> set ?R2) \<subset> unsolvedVars (set ?L1 \<union> set ?L2)" by auto 
      hence "m1 (?R1,?R2) < m1 (?L1,?L2)" by (rule m1psubset)
      thus ?thesis by (rule measures_less)
      
    next

      case False
	with subset have 
	  mem: "(Var x,t) \<in> set (unifs @ seen)" and 
	  eq:  "unsolvedVars ?SUSX = unsolvedVars ?XUS" and
	  xs: "x \<notin> unsolvedVars ?SUSX"
	  by auto
      have eq1: "?SUSX = set (simpleSubstList x t (unifs @ seen) @ [(Var x, t)]) \<union> set []" (is "_ = set (?R1) \<union> set (?R2)") by simp
      have eq2: "?XUS = set ((Var x,t) # unifs) \<union> set seen" (is "_ = set(?L1) \<union> set(?L2)") by simp
      from eq1 and eq2 and subset have "unsolvedVars (set ?R1 \<union> set ?R2) \<subseteq> unsolvedVars (set ?L1 \<union> set ?L2)" by auto 
      hence m1: "m1 (?R1,?R2) \<le> m1 (?L1,?L2)" by (rule m1subset)
      thus ?thesis 
      proof (rule measures_lesseq)
	let ?C = "\<lambda> p S. is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars S"
	show "((?R1,?R2),(?L1,?L2)) \<in> measures [m1a,m2,m3,m4]"
	proof (rule measures_less, unfold m1a_def) 
	  from xs and tnx have xt1: "length [p \<leftarrow> [(Var x,t)]. ?C p ?SUSX] = 1" by simp 
	  have "(\<lambda>(l1, l2). length [p\<leftarrow>l1 @ l2 . ?C p (set l1 \<union> set l2)]) (simpleSubstList x t (unifs @ seen) @ [(Var x, t)], []) 
	    = (length [p\<leftarrow> simpleSubstList x t (unifs @ seen). ?C p ?SUSX] + length [p\<leftarrow> [(Var x, t)]. ?C p ?SUSX])" (is "?RIGHT = ?RR + _")  by simp
	  with xt1 have right: "?RIGHT = Suc ?RR" by auto

	  from eq have "(\<lambda>(l1, l2). length [p\<leftarrow>l1 @ l2 . ?C p (set l1 \<union> set l2)]) ((Var x, t) # unifs, seen) = 
	    length [p\<leftarrow> unifs @ seen. ?C p ?SUSX] + length [p \<leftarrow> [(Var x,t)]. ?C p ?SUSX]" (is "?LEFT = ?LL + _") by simp
	  with xt1 have "?LEFT = ?LL + 1" by (rule subst)
	  hence left: "?LEFT = Suc ?LL" by simp	  
	  
	  have invert: "!! p S. x \<notin> unsolvedVars S \<and> ?C (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t}) S \<Longrightarrow> ?C p S"
	  proof -
	    fix p S
	    assume "x \<notin> unsolvedVars S \<and> ?C (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t}) S" 	    
	    hence xus: "x \<notin> unsolvedVars S" and var: "is_var (fst p \<cdot> {x \<leftarrow> t})" and ne: "fst p \<cdot> {x \<leftarrow> t} \<noteq> snd p \<cdot> {x \<leftarrow> t}" and unsS: "the_var (fst p \<cdot> {x \<leftarrow> t}) \<notin> unsolvedVars S" by auto
	    from ne have ne2: "fst p \<noteq> snd p" by auto
	    show "?C p S"
	    proof (cases "fst p")
	      case Fun with var show ?thesis by (simp add: subst_def)
	    next
	      case (Var y)
	      thus ?thesis 
	      proof (cases "x = y")
		case False 
		with Var have "fst p \<cdot> {x\<leftarrow>t} = Var y" by (unfold subst_def, simp) 
		with unsS and Var and ne2 show "?C p S" by simp
	      next
		case True
		with xus and Var and ne2 show ?thesis by simp
	      qed
	    qed
	  qed

	  have count: "!! L S. ((x \<notin> unsolvedVars S) \<Longrightarrow> length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) L))
	    \<le> length (filter (\<lambda> p. ?C p S) L))"
	  proof  -
	    fix L S
	    assume ass: "x \<notin> unsolvedVars S"
	    show "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) L))
	    \<le> length (filter (\<lambda> p. ?C p S) L)" (is "?INEQ L S")
	    proof (induct L)
	      case Nil thus ?case by simp
	    next
	      case (Cons p ps)
	      show ?case 
	      proof (cases "?C (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t}) S")
		case False
		hence "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (Cons p ps))) = length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ps))" 
		  (is "?SU = _") by simp
		with Cons have "?SU \<le>  length [p\<leftarrow>ps . ?C p S]" by simp
		thus ?thesis by auto 
	      next
		case True
		hence "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (Cons p ps))) = 1+ length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ps))" 
		  (is "?SU = _") by simp
		with Cons have nearly: "?SU \<le>  1+ length [p\<leftarrow>ps . ?C p S]" by auto
		from True have "fst p \<noteq> snd p" by auto
		from ass and True and invert have "?C p S" by blast
		with nearly show ?thesis by simp
	      qed
	    qed
	  qed

	  have count2: "!! (L :: (('f,'v) term \<times> ('f,'v) term) list) (S :: (('f,'v) term \<times> ('f,'v) term) set). \<lbrakk>x \<notin> unsolvedVars S; (Var x, t) mem L\<rbrakk> \<Longrightarrow> length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) L))
	    < length (filter (\<lambda> p. ?C p S) L)"
	  proof -
	    fix L :: "(('f,'v) term \<times> ('f,'v) term) list" and S :: "(('f,'v) term \<times> ('f,'v) term) set"
	    assume ass: "x \<notin> unsolvedVars S" and mem: "(Var x, t) mem L"
	    show "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) L))
	      < length (filter (\<lambda> p. ?C p S) L)" (is "?INEQ L S") using mem
	    proof (induct L)
	      case Nil thus ?case by simp
	    next
	      case (Cons p ps)
	      show ?case 
	      proof (cases "p = (Var x, t)")
		case False 
		with Cons have ind: "(Var x,t) mem ps" by simp
		show ?thesis 
		proof (cases "?C (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t}) S")
		  case False
		  hence "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (Cons p ps))) = length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ps))" 
		    (is "?SU = _") by simp
		  with Cons and ind have "?SU < length [p\<leftarrow>ps . ?C p S]" by simp
		  thus ?thesis by auto
		next
		  case True
		  hence "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (Cons p ps))) = 1+ length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ps))" 
		    (is "?SU = _") by simp
		  with Cons and ind have nearly: "?SU <  1+ length [p\<leftarrow>ps . ?C p S]" by simp
		  from ass and True and invert have "?C p S" by blast
		  with nearly show ?thesis by simp
		qed
	      next
		case True
		from True and ass and tnx have condO: "?C p S" by simp	
		from ass have ind: "length (filter (\<lambda> p. ?C p S) (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ps)) \<le> length (filter (\<lambda> p. ?C p S) ps)" by (rule count)
		have not: "\<not> ?C (t,t) S" by simp 	      
		from xnt have "t \<cdot> {x\<leftarrow>t} = t" by (rule substIdent2)
		with True and not have "\<not> ?C (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t}) S" by (simp add: subst_def)
		with ind and condO show ?thesis by auto 
	      qed
	    qed	  
	  qed

	  
	  have rr: "?RR = length [p\<leftarrow>map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (unifs @ seen)  . is_var (fst p) \<and> fst p \<noteq> snd p \<and> the_var (fst p) \<notin> unsolvedVars ?SUSX]" (is "_ = ?RR2") 
	    by (simp add: simpleSubstListSound)
	  from mem have "(Var x,t) mem (unifs @ seen)" by (simp add: mem_iff)
	  with xs have "?RR2 < ?LL" by (rule count2)
	  with left and right and rr show "?RIGHT < ?LEFT" by arith
	qed
      qed
    qed
  qed
qed

hide const m1 m1a m2 m3 m4	  
	


(* before proving soundness, first have some better induction scheme *) 

(* we first need an ugly one with \<forall> due to the types ! *)

lemma fixes P :: "(('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option \<Rightarrow> bool"
  and res :: "(('f,'v) term \<times> ('f,'v) term) list option" 
  and xs :: "(('f,'v) term \<times> ('f,'v) term) list"
  and ys :: "(('f,'v) term \<times> ('f,'v) term) list"
assumes  
  "(\<forall> ys . res = Some ys \<longrightarrow> P [] ys res)    \<and>
  (\<forall> f ss g ts zs ys . (P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts) \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)  \<and>
  (\<forall> f ss g ts zs ys . res = None \<and> decompose (Fun f ss) (Fun g ts) = None  \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)    \<and>
  (\<forall> f ss x zs ys . (P ((Var x, Fun f ss) # zs) ys res) \<longrightarrow> P ((Fun f ss, Var x) # zs) ys res)    \<and>
  (\<forall> x zs ys . P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res)    \<and>
  (\<forall> x t zs ys .  (res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t) \<longrightarrow> P ((Var x, t) # zs) ys res)    \<and>
  (\<forall> x t zs ys .  (P (map (\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x \<leftarrow> t})) (zs @ ys)  @ [(Var x, t)]) [] res \<and> t \<noteq> Var x \<and> x \<notin> vars_term t) \<longrightarrow> P ((Var x, t) # zs) ys res)    \<and>
  (\<forall> x t zs ys .  (P zs ((Var x,t) # ys) res \<and> t \<noteq> Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (set zs) \<union> varsPairs (set ys)) \<longrightarrow> P ((Var x, t) # zs) ys res)"
  shows unify_pre_induct[rule_format] : "unify xs ys = res \<longrightarrow> P xs ys res"
proof (induct xs ys rule: unify.induct)
  case 1 thus ?case using assms by auto
next
  case (2 f ss g ts unifs simp) 
  thus ?case 
  proof (cases "decompose (Fun f ss) (Fun g ts)")
    case None with 2 assms show ?thesis by auto
  next
    case (Some pairs)
    with 2 have "unify (pairs @ unifs) simp = res \<longrightarrow> P (pairs @ unifs) simp res" by simp
    with 2 have a: "unify (pairs @ unifs) simp = res \<longrightarrow> P (pairs @ unifs) simp res" by simp
    from 2 assms have b: "(\<forall>f ss g ts zs ys . P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)" by simp
    from Some have pz: "pairs = zip ss ts" by (rule decomposeZip)
    from Some have len: "length ss = length ts" by (rule decomposeLength)
    from Some have fg: "f = g" by (rule decomposeFg)
    from Some a b pz len fg show ?thesis by simp
  qed
next
  case (3 f ss x unifs seen) thus ?case using assms by simp
next
  case (4 x t unifs seen)
  thus ?case
  proof (cases "t = Var x")
    case True
    with 4 have imp: "unify unifs seen = res \<longrightarrow> P unifs seen res" by auto
    from 4 assms have one: "(\<forall>x zs ys. P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res)" by simp
    from True one have "P unifs seen res \<longrightarrow> P ((Var x, t) # unifs) seen res" by auto
    with imp True show ?thesis by auto
  next
    case False note oFalse = this
    hence nElem: "t \<noteq> Var x" .
    show ?thesis 
    proof (cases "x mem vars_term_list t")
      case True
      hence xvt: "x \<in>  vars_term t" using vars_term_list_sound by (auto simp: mem_iff)
      from  4 assms have "(\<forall>x t zs ys . res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res)" by simp
      with xvt and nElem True show ?thesis by auto
    next
      case False
      hence xnt: "x \<notin> vars_term t \<and> \<not> x mem vars_term_list t" using vars_term_list_sound by (auto simp: mem_iff)
      show ?thesis
      proof (cases "varsL (unifs @ seen) x")
	case False
	hence xvp: "x \<notin> varsPairs (set unifs \<union> set seen)" by simp
	with 4 assms nElem xnt have "P unifs ((Var x, t) # seen) res \<longrightarrow> P ((Var x, t) # unifs) seen res" by simp
	with 4 assms 
	  and xnt and nElem and False show ?thesis by auto
      next
	case True
	with 4 nElem xnt assms have imp: "P (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (unifs @ seen) @ [(Var x, t)]) [] res  \<longrightarrow>
      P ((Var x, t) # unifs) seen res" by simp
	from 4
	  and nElem and xnt and True have
	  "unify (simpleSubstList x t (unifs @ seen) @ [(Var x, t)]) [] = res 
	  \<longrightarrow> P (simpleSubstList x t (unifs @ seen) @ [(Var x, t)]) [] res"  
	  by simp
	with imp and nElem and xnt and True show ?thesis by (simp add: simpleSubstListSound)
      qed
    qed
  qed 
qed 
	  
(* now here is the induction scheme you should use *)

lemma  
  assumes Done: "\<And>ys. res = Some ys \<Longrightarrow> P [] ys res" 
  and FF: "\<And>f ss g ts zs ys. \<lbrakk>P (zip ss ts @ zs) ys res; decompose (Fun f ss) (Fun g ts) = Some (zip ss ts); f = g; length ss = length ts\<rbrakk> \<Longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res"
  and Clash: "\<And>f ss g ts zs ys. \<lbrakk>res = None; decompose (Fun f ss) (Fun g ts) = None\<rbrakk>  \<Longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res"
  and Swap: "\<And>f ss x zs ys. P ((Var x, Fun f ss) # zs) ys res \<Longrightarrow> P ((Fun f ss, Var x) # zs) ys res"
  and XX: "\<And>x zs ys. P zs ys res \<Longrightarrow> P ((Var x, Var x) # zs) ys res"
  and Occur: "\<And>x t zs ys.  \<lbrakk>res = None; Var x \<noteq> t; x \<in> vars_term t\<rbrakk> \<Longrightarrow> P ((Var x, t) # zs) ys res"
  and Subst: "\<And>x t zs ys.  \<lbrakk>P (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys) @ [(Var x, t)]) [] res; t \<noteq> Var x; x \<notin> vars_term t\<rbrakk> \<Longrightarrow> P ((Var x, t) # zs) ys res"
  and Solved: "\<And>x t zs ys.  \<lbrakk>P zs ((Var x,t) # ys) res; t \<noteq> Var x; x \<notin> vars_term t; x \<notin> varsPairs (set zs \<union> set ys)\<rbrakk> \<Longrightarrow> P ((Var x, t) # zs) ys res" 
  shows unify_induct [case_names Done FF Clash Swap XX Occur Subst Solved] : "unify xs ys = res \<longrightarrow> P xs ys res"
proof
  assume ass: "unify xs ys = res" 
  have "(\<forall>ys. res = Some ys \<longrightarrow> P [] ys res) \<and>
 (\<forall>f ss g ts zs ys.
     P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts \<longrightarrow>
     P ((Fun f ss, Fun g ts) # zs) ys res) \<and>
 (\<forall>f ss g ts zs ys. res = None \<and> decompose (Fun f ss) (Fun g ts) = None \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res) \<and>
 (\<forall>f ss x zs ys. P ((Var x, Fun f ss) # zs) ys res \<longrightarrow> P ((Fun f ss, Var x) # zs) ys res) \<and>
 (\<forall>x zs ys. P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res) \<and>
 (\<forall>x t zs ys. res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res) \<and>
 (\<forall>x t zs ys.
     P (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys) @ [(Var x, t)]) [] res \<and>
     t \<noteq> Var x \<and> x \<notin> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res) \<and>
 (\<forall>x t zs ys.
     P zs ((Var x, t) # ys) res \<and> t \<noteq> Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (set zs) \<union> varsPairs (set ys) \<longrightarrow>
     P ((Var x, t) # zs) ys res)" using assms by auto
  from this and ass show "P xs ys res" by (rule unify_pre_induct)
qed




definition solved where "solved pairs == \<forall> (s,t) \<in> pairs. \<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (pairs - {(s,t)})"

lemma solvedMono : assumes sub: "xs \<subseteq> ys" and solve: "solved ys" shows "solved xs"
proof (unfold solved_def, rule ballI)
  fix p
  assume "p \<in> xs"
  show "(\<lambda>(s, t). \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (xs - {(s, t)})) p"
  proof (cases p)
    case (Pair s t)
    with `p \<in> xs` and sub have "(s,t) \<in> ys" by force
    with solve have "\<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (ys - {(s,t)})" by (unfold solved_def, auto)
    from this obtain x where "s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (ys - {(s,t)})" ..
    with sub have "s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (xs - {(s,t)})" by (unfold varsPairs_def, auto)
    with Pair show ?thesis by force 
  qed
qed
  
  

lemma solvedMainProperty : fixes pairs :: "(('f,'v) term \<times> ('f,'v) term) list"
  assumes solved: "solved (set pairs)"
  shows "(subst_of pairs) \<in> unifiers (set pairs) \<and> (\<forall> \<sigma> \<in> unifiers (set pairs). \<sigma> = (subst_of pairs) \<circ>s \<sigma>)" 
proof -
  let ?S = "subst_of pairs"
  
  have "solved (set pairs \<union> set pairs) \<longrightarrow> (\<exists> (f :: 'v \<Rightarrow> ('f,'v) term). ?S = Sub f \<and> (\<forall> p \<in> set pairs. f (the_var (fst p)) = snd p) \<and> 
    (\<forall> \<sigma> \<in> unifiers (set pairs). \<sigma> = ?S \<circ>s \<sigma>) \<and> (dom ?S \<inter> \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {}))"
  proof (induct pairs rule: list.induct[where ?P = "\<lambda> list. solved (set pairs \<union> set list) \<longrightarrow> 
      (\<exists> (f :: 'v \<Rightarrow> ('f,'v) term). subst_of list = Sub f \<and> (\<forall> p \<in> set list. f (the_var (fst p)) = snd p) \<and> 
    (\<forall> \<sigma> \<in> unifiers (set list). \<sigma> = subst_of list \<circ>s \<sigma>) \<and> (dom (subst_of list) \<inter> \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {}))"])
    case Nil thus ?case 
    proof (auto simp: unifiers_def sub_empty_def)
      fix \<sigma> 
      show "\<sigma> = Sub (get_sub \<sigma>)" by (cases "\<sigma>", auto)
    qed
      
  next
    case (Cons p ps)
    show ?case 
    proof
      assume solvepps: "solved (set pairs \<union> set (p # ps))"
      have "set pairs \<union> set ps \<subseteq> set pairs \<union> set (p # ps)" by auto
      from this and solvepps have solveps: "solved (set pairs \<union> set ps)" by (rule solvedMono)
      from this and Cons obtain f where ind: "subst_of ps = Sub f \<and> (\<forall> p \<in> set ps. f (the_var (fst p)) = snd p) \<and> 
    (\<forall> \<sigma> \<in> unifiers (set ps). \<sigma> = subst_of ps \<circ>s \<sigma>) \<and> (dom (subst_of ps) \<inter> \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {})" by blast 
      hence indSub: "subst_of ps = Sub f" and 
        indPs: "\<forall> p \<in> set ps. f (the_var (fst p)) = snd p" and
        indUn: "\<forall> \<sigma> \<in> unifiers (set ps). \<sigma> = subst_of ps \<circ>s \<sigma>" and
        indDom: "dom (subst_of ps) \<inter> \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {}"
	by auto
      show "\<exists>f. subst_of (p # ps) = Sub f \<and>
        (\<forall>p\<in>set (p # ps). f (the_var (fst p)) = snd p) \<and> (\<forall>\<sigma>\<in>unifiers (set (p # ps)). \<sigma> = subst_of (p # ps) \<circ>s \<sigma>) \<and> dom (subst_of (p # ps)) \<inter> \<Union>(vars_term \<circ> snd) ` set pairs \<subseteq> {}"
      proof (cases p)
	case (Pair s t)
	show ?thesis
	proof (cases s)
	  case Fun
	  with Pair and solvepps have False by (unfold solved_def, auto) thus ?thesis ..
	next
	  case (Var x)
	  with Pair have p: "p = (Var x, t)" by simp
	  let ?g = "(\<lambda>y. if x = y then t else f y) :: 'v \<Rightarrow> ('f,'v) term"
	  have "subst_of ((Var x, t) # ps) = Sub ?g"
	  proof (simp)
	    show "(\<lambda>y. if x = y then t else case subst_of ps of Sub s \<Rightarrow> s y) = ?g"
	    proof (rule ext)
	      fix y
	      from indSub show "(if x = y then t else case subst_of ps of Sub s \<Rightarrow> s y) = ?g y" by (cases "x = y", auto)
	    qed
	  qed
	  with p have one: "subst_of (p # ps) = Sub ?g" by simp
	  
	  have two: "\<forall>q\<in>set (p # ps). ?g (the_var (fst q)) = snd q" 
	  proof
	    fix q :: "(('f,'v) term \<times> ('f,'v) term)"
	    assume qpps: "q \<in> set (p # ps)"
	    let ?LL = "(if x = the_var (fst q) then t else f (the_var (fst q))) :: ('f,'v) term"
	    show "?LL = snd q" 
	    proof (cases q)
	      case (Pair u v)
	      show ?thesis 
	      proof (cases "u")
		case Fun with qpps and Pair and solvepps have False by (unfold solved_def, auto) thus ?thesis ..
	      next
		case (Var y)
		with Pair have q: "q = (Var y, v)" by simp
		have "(if x = y then t else f y) = v" 
		proof (cases "x = y")
		  case False 
		  with q and p and `q \<in> set (p # ps)` have "q \<in> set ps" by simp
		  with indPs and q have "f y = v" by auto
		  with False show ?thesis by simp
		next
		  case True
		  show ?thesis 
		  proof (cases "t = v") 
		    case True with `x = y` show ?thesis by simp
		  next
		    case False
		    have "p \<in> set (p # ps)" by simp
		    with p and solvepps have "\<exists>z. Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> varsPairs (set pairs \<union> set (p # ps) - {(Var x, t)})" by (unfold solved_def, auto)
		    from this obtain z where "Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> varsPairs (set pairs \<union> set (p # ps) - {(Var x, t)})" ..
		    hence contra: "x \<notin> varsPairs (set (p # ps) - {(Var x, t)})" by (unfold varsPairs_def, auto)
		    from False and p and q and `q \<in> set (p # ps)` have qset: "q \<in> set (p # ps) - {(Var x,t)}" by blast
		    from True and q have "x \<in> vars_term(fst q)" by simp 
		    with qset have "x \<in> varsPairs (set (p # ps) - {(Var x, t)})"
		      by (unfold varsPairs_def, blast)
		    with contra have False .. thus ?thesis ..
		  qed
		qed		
		with p q show ?thesis by auto
	      qed
	    qed
	  qed

	  have three: "\<forall> \<sigma> \<in> unifiers (set (p # ps)). \<sigma> = subst_of (p # ps) \<circ>s \<sigma>" 
	  proof
	    fix \<sigma> 
	    assume "\<sigma> \<in> unifiers (set (p # ps))" 
	    with p have "\<sigma> \<in> {\<sigma> . Var x\<cdot>\<sigma> = t\<cdot>\<sigma>} \<inter> unifiers (set ps)" by (simp add: unifiersInsert)
	    hence xts: "Var x \<cdot> \<sigma> = t \<cdot> \<sigma>" and su: "\<sigma> \<in> unifiers (set ps)" by auto
	    from su and indUn and indSub have ind: "\<sigma> = Sub f \<circ>s \<sigma>" by simp
	    show "\<sigma> = subst_of (p # ps) \<circ>s \<sigma>"
	    proof (cases \<sigma>)
	      case (Sub sigma)
	      with ind have "Sub sigma = Sub f \<circ>s Sub sigma" by simp
	      have "Sub ?g \<circ>s Sub sigma = Sub sigma" 
	      proof simp
		show "(\<lambda>y. (if x = y then t else f y) \<cdot> Sub sigma) = sigma"
		proof (rule ext)
		  fix y
		  show "(if x = y then t else f y) \<cdot> Sub sigma = sigma y"
		  proof (cases "x = y")
		    case False
		    from Sub and ind have "f y \<cdot> Sub sigma = sigma y" proof -
		      have "f y \<cdot> Sub sigma = Var y \<cdot> Sub f \<cdot> Sub sigma" by simp
		      also have "\<dots> = Var y \<cdot> Sub f \<circ>s Sub sigma" by simp
		      also have "\<dots> = Var y \<cdot> Sub sigma" using ind Sub by simp
		      also have "\<dots> = sigma y" by simp
		      finally show ?thesis . 
		    qed
		    with False show ?thesis by auto
		  next
		    case True
		    with xts and Sub show ?thesis by simp 		    
		  qed
		qed
	      qed
	      with one and Sub show ?thesis by simp
	    qed
	  qed

	  have four : "dom (subst_of (p # ps)) \<inter> \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {}" 
	  proof
	    fix y
	    assume "y \<in> dom (subst_of (p # ps)) \<inter> \<Union>(vars_term \<circ> snd) ` set pairs"
	    hence ypps: "y \<in> dom (subst_of (p # ps))" and ypairs: "y \<in> \<Union> (vars_term \<circ> snd) ` set pairs" by auto
	    from ypps and one have "y \<in> dom (Sub ?g)" by simp
	    with p have "y = x \<or> y \<in> dom (Sub f)"
	      by (cases "y = x", auto)
	    with indSub have "y = x \<or> y \<in> dom (subst_of ps)" by simp
	    with indDom and ypairs have yx: "y = x" by auto
	    from this and ypairs obtain q where qpairs: "q \<in> set pairs \<union> set (p # ps)" and xq: "x \<in> vars_term(snd q)" by force
	    from p have  "(Var x, t) \<in> set pairs \<union> set (p # ps)" by simp
	    from this and solvepps have "\<exists>z. Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> varsPairs (set pairs \<union> set (p # ps) - {(Var x, t)})" 
	      by (unfold solved_def, blast)
	    from this have xt: "x \<notin> vars_term t" and xnmem: "x \<notin> varsPairs (set pairs \<union> set (p # ps) - {(Var x,t)})" by auto

	    have False
	    proof (cases q)
	      case (Pair u v)
	      with xq have xv: "x \<in> vars_term v" by simp
	      from Pair and qpairs have uvpair: "(u,v) \<in> set pairs \<union> set (p # ps)" by simp
	      with solvepps have "\<exists>z. u = Var z \<and> z \<notin> vars_term v \<and> z \<notin> varsPairs (set pairs \<union> set (p # ps) - {(u, v)})" by (unfold solved_def, blast)
	      from this obtain z where uz: "u = Var z" and  zv: "z \<notin> vars_term v" by blast
	      from zv and Pair and xq have "x \<noteq> z" by auto
	      with uvpair and uz have uvpair2: "(u,v) \<in> set pairs \<union> set (p # ps) - {(Var x, t)}" by blast
	      from xv have "x \<in> vars_term(snd (u, v))" by simp
	      with uvpair2 have "x \<in> varsPairs (set pairs \<union> set (p # ps) - {(Var x,t)})" by (unfold varsPairs_def, blast)
	      with xnmem show False ..
	    qed
	    thus "y \<in> {}" ..
	  qed	  

	  from one two three four show ?thesis by blast
	qed
      qed
    qed
  qed

  with solved obtain f where 
    one: "?S = Sub f" and 
    two: "\<forall>p\<in>set pairs. f (the_var (fst p)) = snd p" and 
    three: "\<forall>\<sigma>\<in>unifiers (set pairs). \<sigma> = ?S \<circ>s \<sigma>" and 
    four: "dom ?S \<inter>  \<Union> ((vars_term \<circ> snd) ` set pairs) \<subseteq> {}" by force
  show ?thesis
  proof 
    from one and two show "?S \<in> unifiers (set pairs)"
    proof (unfold unifiers_def, auto)
      fix s t
      assume "(s,t) \<in> set pairs"
      show "s \<cdot> Sub f = t \<cdot> Sub f"
      proof (cases s)
	case Fun with `solved (set pairs)` and `(s,t) \<in> set pairs` have False by (unfold solved_def, auto) thus ?thesis ..
      next
	case (Var x)
	with two and `(s,t) \<in> set pairs` and two have piece1: "s \<cdot> Sub f = t \<cdot> \<iota>" by auto
	from four and `(s,t) \<in> set pairs` have "dom ?S \<inter> vars_term t = {}" by force
	with one have noInter: "vars_term t \<inter> dom (Sub f) = {}" by auto
	have "finite (vars_term t)" by simp
	from this and noInter have piece2: "(Sub f)|_(vars_term t) = \<iota>" by (rule restrict_empty_set)
	have piece3: "t \<cdot> (Sub f) = t \<cdot> ((Sub f)|_(vars_term t))" by (rule coincidence_lemma)
	from piece1 piece2 piece3  show ?thesis by simp
      qed
    qed
    
  next
    
    from three show "\<forall> \<sigma> \<in> unifiers (set pairs). \<sigma> = subst_of pairs \<circ>s \<sigma>" .
    
  qed
qed

lemma solvedSingle : assumes xnt: "x \<notin> vars_term t" shows "{x\<leftarrow>t} \<in> unifiers {(Var x, t)} \<and> (\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = {x\<leftarrow>t} \<circ>s \<sigma>)"
proof -
  from xnt have solve: "solved (set [(Var x,t)])" (is "solved ?P") by (unfold solved_def varsPairs_def, auto)
  have eq: "{x\<leftarrow>t} = subst_of [(Var x,t)]" (is "?XT = ?S") by (unfold subst_def, simp, rule ext, force simp: sub_empty_def)
  from solve have "?S \<in> unifiers ?P \<and> (\<forall> \<sigma> \<in> unifiers ?P. \<sigma> = ?S  \<circ>s \<sigma>)"
    by (rule solvedMainProperty)
  with eq show ?thesis by simp
qed
  
  
lemma substSound: assumes xnt: "x \<notin> vars_term t" shows "unifiers (insert (Var x, t) s) = unifiers (insert (Var x,t) ((\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` s))"
proof -
  from xnt have "{x\<leftarrow>t} \<in> unifiers {(Var x, t)} \<and> (\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = {x\<leftarrow>t} \<circ>s \<sigma>)" by (rule solvedSingle)
  hence general: "\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = {x\<leftarrow>t} \<circ>s \<sigma>" ..
  show ?thesis 
  proof (rule set_ext)
    fix \<theta> 
    have "(\<theta> \<in> unifiers (insert (Var x, t) s)) = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> \<theta> \<in> unifiers s)" by (simp add: unifiersInsert)
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> {x\<leftarrow>t} \<circ>s \<theta> \<in> unifiers s)" 
    proof -
      have "Var x \<cdot> \<theta> = t \<cdot> \<theta> \<Longrightarrow> \<theta> \<in> unifiers {(Var x,t)}" by (simp add: unifiersInsert)
      with general have "Var x \<cdot> \<theta> = t \<cdot> \<theta> \<Longrightarrow> \<theta> = {x\<leftarrow>t} \<circ>s \<theta>" by blast
      thus ?thesis by force
    qed
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> p \<in> s. fst p \<cdot> {x\<leftarrow>t} \<circ>s \<theta> = snd p \<cdot> {x\<leftarrow>t} \<circ>s \<theta>))" by (simp add: unifiers_def)
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> p \<in> s. fst p \<cdot> {x\<leftarrow>t} \<cdot> \<theta> = snd p \<cdot> {x\<leftarrow>t} \<cdot> \<theta>))" by force
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> q \<in> ((\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` s). fst q \<cdot> \<theta> = snd q \<cdot> \<theta>))" by force
    also have "\<dots> = (\<theta> \<in> unifiers (insert (Var x,t) ((\<lambda> p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` s)))" by (simp add:  unifiers_def)
    finally show "(\<theta> \<in> unifiers (insert (Var x, t) s)) = (\<theta> \<in> unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` s)))" .
  qed
qed



(* result None shows that there are no unifiers *)
lemma unifySound1 : assumes "unify xs ys = None" (is "_ = ?R") shows "unifiers (set xs \<union> set ys) = {}" (is ?Q)
proof -
  have "unify xs ys = ?R \<longrightarrow> ?Q"
  proof (induct rule: unify_induct)
    case Done thus ?case by simp
  next
    case FF thus ?case by auto 
  next
    case (Clash f ss g ts)
    thus ?case by simp
  next 
    case Swap thus ?case by simp
  next
    case XX thus ?case by simp
  next
    case Occur thus ?case by simp
  next
    case (Subst x t zs ys)
    have "(set (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys) @ [(Var x, t)]) \<union> set []) = 
      insert (Var x, t) (set (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys)))" (is "?L = _") by simp
    also have "\<dots> = insert (Var x, t) ((\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` (set (zs @ ys)))" (is "_ = ?R") by (auto)
    finally have eq:  "unifiers ?L = unifiers ?R" by simp
    
    have "unifiers (set ((Var x, t) # zs) \<union> set ys) = unifiers (insert (Var x, t) (set (zs @ ys)))" by simp
    also have "\<dots> = unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` (set (zs @ ys))))" using `x \<notin> vars_term t` by (rule substSound)
    also have "\<dots> = {}" using eq and Subst by simp
    finally show ?case .
  next
    case Solved thus ?case by simp
  qed
  with assms show ?thesis by simp
qed 

(* result Some pairs has the same unifiers as input *)
lemma unifySound2 : assumes "unify xs ys = Some us" (is "_ = ?R") shows "unifiers (set xs \<union> set ys) = unifiers (set us)" (is ?Q)
proof -
  have "unify xs ys = ?R \<longrightarrow> ?Q"
  proof (induct rule: unify_induct)
    case Done thus ?case by simp
  next
    case FF thus ?case by auto
  next
    case Clash thus ?case by simp
  next 
    case Swap thus ?case by simp
  next
    case XX thus ?case by simp
  next
    case Occur thus ?case by auto
  next
    case (Subst x t zs ys) 
    have "(set (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys) @ [(Var x, t)]) \<union> set []) = 
      insert (Var x, t) (set (map (\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) (zs @ ys)))" (is "?L = _") by simp
    also have "\<dots> = insert (Var x, t) ((\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` (set (zs @ ys)))" (is "_ = ?R") by (auto)
    finally have eq:  "unifiers ?L = unifiers ?R" by simp
    
    have "unifiers (set ((Var x, t) # zs) \<union> set ys) = unifiers (insert (Var x, t) (set (zs @ ys)))" by simp
    also have "\<dots> = unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> {x\<leftarrow>t}, snd p \<cdot> {x\<leftarrow>t})) ` (set (zs @ ys))))" using `x \<notin> vars_term t` by (rule substSound)
    also have "\<dots> = unifiers (set us)" using eq and Subst by simp
    finally show ?case .

  next
    case Solved thus ?case by simp
  qed
  with assms show ?thesis by simp
qed 
 
    

(* solved compution *) 
definition solvedC where "solvedC seen todo == \<forall> (s,t) \<in> seen. \<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs ((seen - {(s,t)}) \<union> todo)"

(* if everything in seen is in solved form, then the result is in solved form *)  
lemma unifySound3 :  
  assumes "unify xs ys = Some us" and "solvedC (set ys) (set xs)" 
  shows "solvedC (set us) {}" 
proof -
  have "unify xs ys = Some us \<longrightarrow> (solvedC (set ys) (set xs) \<longrightarrow> solvedC (set us) {})"
  proof (induct rule: unify_induct[where ?xs = "xs" and ?ys = "ys"])
    case Done thus ?case by simp
  next
    case Clash thus ?case by simp
  next
    case Occur thus ?case by auto
  next
    case (Swap f ss x zs) thus ?case by (auto simp: solvedC_def)
  next
    case XX thus ?case by (auto simp: solvedC_def)
  next
    case Subst thus ?case by (unfold solvedC_def, simp)
  next
    case (FF f ss g ts zs ys) 
    let ?Y = "set ys"
    let ?F = "{(Fun f ss, Fun g ts)}"
    let ?Z = "set zs"
    let ?ZIP = "set (zip ss ts)"
    have "solvedC ?Y (?F \<union> ?Z) \<Longrightarrow> solvedC ?Y (?ZIP \<union> ?Z)"
    proof (unfold solvedC_def)
      assume ind: "\<forall>(s, t) \<in> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?F \<union> ?Z))"
      show "\<forall>(s, t) \<in> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))"
      proof 
	fix p
	assume "p \<in> ?Y"
	show "(\<lambda>(s, t). \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))) p"
	proof (cases p)
	  case (Pair s t)
	  with `p \<in> set ys` and ind have "\<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?F \<union> ?Z))" by auto
	  from this obtain x where c1: "s = Var x \<and> x \<notin> vars_term t" and xF: "x \<notin> varsPairs (?Y - {(s, t)} \<union> (?F \<union> ?Z))" by auto
	  from xF have easy: "x \<notin> varsPairs (?Y - {(s,t)}) \<and> x \<notin> varsPairs ?Z" and f: "x \<notin> vars_term(Fun f ss) \<union> vars_term(Fun g ts)" by auto
	  have "x \<notin> varsPairs ?ZIP" 
	  proof 
	    assume "x \<in> varsPairs ?ZIP"
	    hence "x \<in> vars_term(Fun f ss) \<or> x \<in> vars_term(Fun g ts)" by (rule varsPairsFun)
	    with f show False by simp
	  qed
	  with easy have "x \<notin> varsPairs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))" by simp
	  with c1 have "\<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))" by blast
	  with Pair show ?thesis by simp
	qed
      qed
    qed
    with FF show ?case by auto
  next
    case (Solved y u zs ys) 
    let ?Y = "set ys"
    let ?X = "{(Var y, u)}"
    let ?Z = "set zs"
    have "solvedC ?Y (?X \<union> ?Z) \<Longrightarrow> solvedC (?X \<union> ?Y) ?Z" 
    proof (unfold solvedC_def)
      assume ind: "\<forall>(s, t) \<in> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s,t)} \<union> (?X \<union> ?Z))"
      show "\<forall>(s, t) \<in> ?X \<union> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?X \<union> ?Y - {(s,t)} \<union> ?Z)"
      proof 
	fix p
	assume "p \<in> ?X \<union> ?Y"
	hence pXY: "(p \<in> ?X \<and> p \<notin> ?Y) \<or> p \<in> ?Y" by blast
	show "(\<lambda>(s, t). \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?X \<union> ?Y - {(s, t)} \<union> ?Z)) p"
	proof (cases p)
	  case (Pair s t)
	  from pXY show ?thesis 
	  proof 
	    assume "p \<in> ?Y"
	    with ind and Pair have "\<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> varsPairs (?Y - {(s, t)} \<union> (?X \<union> ?Z))" by blast
	    from this obtain x where c1: "s = Var x \<and> x \<notin> vars_term t" and xOld: "x \<notin> varsPairs (?Y - {(s, t)} \<union> (?X \<union> ?Z))" by auto
	    from xOld have "x \<notin> varsPairs (?X \<union> ?Y - {(s, t)} \<union> ?Z)" by (unfold varsPairs_def, blast)
	    with Pair and c1 show ?thesis by simp  
	  next
	    assume "p \<in> ?X \<and> p \<notin> ?Y"
	    hence pX: "p \<in> ?X" and pY: "p \<notin> ?Y" by auto
	    from pX and Pair have sy: "s = Var y" and tu: "t = u" by auto
	    have "y \<notin> varsPairs (?X \<union> ?Y - {(s,t)} \<union> ?Z)"
	    proof
	      assume "y \<in> varsPairs (?X \<union> ?Y - {(s,t)} \<union> ?Z)"
	      with sy tu Solved show False  by (unfold varsPairs_def, blast)
	    qed
	    with sy and tu and Pair `y \<notin> vars_term u` show ?thesis by simp 
	  qed
	qed
      qed 
    qed
    with Solved show ?case by simp
  qed
  with assms show ?thesis by simp
qed


(* when starting the unification algorithm properly, then the result is in solved form *)	  
lemma unifySound4 : assumes res: "unify pairs [] = Some res" shows "solved (set res)"
proof -
  have "solvedC (set []) (set pairs)" by (unfold solvedC_def, simp)
  with  res have "solvedC (set res) {}" by (rule unifySound3)
  thus ?thesis by (unfold solved_def solvedC_def, auto)
qed


lemma mguComplete : "mgu s t = None \<Longrightarrow> unifiers {(s,t)} = {}"
proof -
  assume "mgu s t = None" 
  hence "unify [(s,t)] [] = None" by (cases "unify [(s,t)] []", auto)
  hence "unifiers (set [(s,t)] \<union> set []) = {}" by (rule unifySound1)
  thus ?thesis by simp
qed

lemma mguSound : "mgu s t = Some \<delta> \<Longrightarrow> isSMgu \<delta> {(s,t)}"
proof -
  assume result: "mgu s t = Some \<delta>"
  show ?thesis 
  proof (cases "unify [(s,t)] []")
    case None with result show ?thesis by simp
  next
    case (Some res) 
    hence "unifiers (set [(s,t)] \<union> set []) = unifiers (set res)" by (rule unifySound2)
    hence unif: "unifiers {(s,t)} = unifiers (set res)" by simp 
    from Some have "solved (set res)" by (rule unifySound4)
    hence "subst_of res \<in> unifiers (set res) \<and> (\<forall>\<sigma>\<in>unifiers (set res). \<sigma> = subst_of res \<circ>s \<sigma>)" by (rule solvedMainProperty)
    with unif and result and Some have "\<delta> \<in> unifiers {(s,t)} \<and> (\<forall>\<sigma>\<in>unifiers {(s,t)}. \<sigma> = \<delta> \<circ>s \<sigma>)" by force
    thus ?thesis by (unfold isSMgu_def, force)
  qed
qed 


lemma mgu_of[simp] :  assumes unif: "s \<cdot> \<sigma> = t \<cdot> \<sigma>" shows "s \<cdot> mgu_of s t = t \<cdot> mgu_of s t \<and> \<sigma> = mgu_of s t \<circ>s \<sigma>" 
proof -
  from unif have ne: "\<sigma> \<in> unifiers {(s,t)}" by (unfold unifiers_def, force)
  show ?thesis
  proof (cases "mgu s t")
    assume "mgu s t = None"
    hence "unifiers {(s,t)} = {}" by (rule mguComplete)
    with ne have False by simp thus ?thesis ..
  next
    fix \<delta> 
    assume res: "mgu s t = Some \<delta>"
    hence "isSMgu \<delta> {(s,t)}" by (rule mguSound)
    with res have "isSMgu (mgu_of s t) {(s,t)}" by (unfold mgu_of_def, simp)
    with ne show ?thesis by (unfold isSMgu_def unifiers_def, auto)
  qed
qed



(* a sound and complete matching algorithm *)

definition matcher :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)sub set"
  where "matcher t l == {\<sigma>. t = l \<cdot> \<sigma>}"

definition matcherL :: "(('f,'v)term \<times> ('f,'v)term)set \<Rightarrow> ('f,'v)sub set"
   where "matcherL ps == {\<sigma>. (\<forall> p \<in> ps. fst p = snd p \<cdot> \<sigma>)}"
abbreviation matcherLi :: "('v \<times> ('f,'v)term) set \<Rightarrow> ('f,'v)sub set"
    where "matcherLi ps == matcherL ((\<lambda> p. (snd p, Var (fst p))) ` ps)" 

function matchList :: "(('f,'v)term \<times> ('f,'v)term)list \<Rightarrow> ('v \<times> ('f,'v)term) list option"
  where "matchList [] = Some []"
    |   "matchList ((Var _, Fun _ _) # ps) = None"
    |   "matchList ((Fun f ts,Fun g ls) # ps) = (if f = g then 
             (case decomp ts ls of Some pairs \<Rightarrow> matchList (pairs @ ps) | None \<Rightarrow> None) else None)"
    |   "matchList ((t, Var x) # ps) = 
             (case matchList ps of Some pairs \<Rightarrow> Some ((x,t) # pairs) | None \<Rightarrow> None)"
by pat_completeness auto
termination 
proof -
  let ?M = "measures [\<lambda>l. listsum (map (sz \<circ> snd) l), length]"
  show ?thesis 
  proof
    show "wf ?M" by auto    
  next
    fix t x ps 
    show "(ps, (t, Var x) # ps) \<in> ?M" by auto
  next
    fix f :: 'f and ts :: "('f,'v)term list" and g :: 'f and ls :: "('f,'v)term list" and ps res
    assume dec: "decomp ts ls = Some res"
    show "(res @ ps, (Fun f ts, Fun g ls) # ps) \<in> ?M" 
    proof (rule measures_less, auto)
      from dec have len: "length ts = length ls" by (rule decompLength)
      from dec have res: "res = zip ts ls" by (rule decompZip)
      from len and res show "listsum (map (sz \<circ> snd) res) < Suc (listsum (map sz ls))" by (auto simp: map_map[symmetric])
    qed
  qed
qed


lemma matcherLUnion[simp] : "matcherL (ps \<union> qs) = matcherL ps \<inter> matcherL qs"
  by (unfold matcherL_def, blast)

lemma matcherLInsert : "matcherL (insert p ps) = {\<sigma>. fst p = snd p \<cdot> \<sigma>} \<inter> matcherL ps"
  by (unfold matcherL_def, blast)

lemma matcherLZip : assumes "length ts = length ls" shows 
  "{\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>} = matcherL (set (zip ts ls))" 
using assms
proof (induct ls arbitrary:ts) 
  case Nil thus ?case by (cases ts, auto simp:matcherL_def)
next
  case (Cons l ls) note oCons=this
  thus ?case 
  proof (cases ts)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons t tts)
    from Cons oCons have "length tts = length ls" by simp
    with Cons oCons have Prop: "{\<sigma>. \<forall>i<length tts. tts ! i = ls ! i \<cdot> \<sigma>} = matcherL (set (zip tts ls))" by simp
    have "{\<sigma>. \<forall>i<length (t # tts). (t # tts) ! i = (l # ls) ! i \<cdot> \<sigma>} = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> {\<sigma>. \<forall>i<length tts. tts ! i = ls ! i \<cdot> \<sigma>}" (is "?l = ?r")
    proof (rule set_ext, rule iffI)
      fix s
      assume "s \<in> ?l"
      thus "s  \<in> ?r" by force
    next
      fix s 
      assume Prop2: "s \<in> ?r"
      hence one: "t = l \<cdot> s" by simp
      from Prop2 have two: "\<forall>i< length tts. tts ! i = ls ! i \<cdot> s" by simp
      have "\<forall> i < length (t # tts). (t # tts) ! i = (l # ls) ! i \<cdot> s"
      proof (rule allI, rule impI)
	fix i
	assume "i < length (t # tts)"
	with one two show "(t # tts) ! i = (l # ls) ! i \<cdot> s"
	  by (cases i, auto)
      qed
      thus "s \<in> ?l" by simp
    qed
    also have "\<dots> = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> matcherL (set (zip tts ls))" using Prop by simp
    finally show ?thesis using Cons by (simp add: matcherLInsert)
  qed
qed

lemma matcherLFun[simp] : "matcherL (insert (Fun f ts, Fun g ls) ps) = (if f = g \<and> length ts = length ls then matcherL (set (zip ts ls)) \<inter> matcherL ps else
{})" 
proof (cases "f = g \<and> length ts = length ls")
  assume "\<not> (f = g \<and> length ts = length ls)"
  hence "f \<noteq> g \<or> length ts \<noteq> length ls" by blast
  thus ?thesis
  proof
    assume "f \<noteq> g" thus ?thesis by (unfold matcherL_def, simp)
  next
    assume "length ts \<noteq> length ls" thus ?thesis by (unfold matcherL_def, auto)
  qed
next
  assume P: "f = g \<and> length ts = length ls"
  hence "{\<sigma>. ts = map (\<lambda> l. l \<cdot> \<sigma>) ls} = {\<sigma>. \<forall>i<length ts. ts ! i = (map (\<lambda> l. l \<cdot> \<sigma>) ls) ! i}"  using list_eq_iff_nth_eq[where xs = ts] by force
  also have "\<dots> = {\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>}" using P by auto 
  also have "\<dots> = matcherL (set (zip ts ls))"
    using P matcherLZip by blast
  finally show ?thesis using P by (simp add: matcherLInsert)
qed
 

lemma matchListComplete : assumes "matchList ps = None" shows "matcherL (set ps) = {}"
using assms 
proof (induct rule:matchList.induct)
  case 1 thus ?case by simp
next
  case (2 x f ls ps) thus ?case by (unfold matcherL_def, auto)
next
  case (4 t x ps) thus ?case by (cases "matchList ps", unfold matcherL_def, auto)
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False thus ?thesis by (unfold matcherL_def, auto)
  next
    case True thus ?thesis 
    proof (cases "decomp ts ls")
      case (Some pairs)
      hence len: "length ts = length ls" by (rule decompLength)
      from Some have zip: "pairs = zip ts ls" by (rule decompZip)
      with Some `f = g` 3 have "matcherL (set (zip ts ls @ ps)) = {}"  by simp
      with len show ?thesis by simp
    next
      case None
      hence difflen: "length ts \<noteq> length ls" by (rule decompNone)
      thus ?thesis by simp
    qed
  qed
qed

lemma matchListSound : assumes "matchList ps = Some qs" shows "matcherL (set ps) = matcherLi (set qs)" 
using assms 
proof (induct arbitrary:qs rule:matchList.induct)
  case 1 thus ?case by (unfold matcherL_def, simp)
next
  case (2 x f ls ps) thus ?case by auto
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis 
    proof (cases "decomp ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some pairs)
      hence len: "length ts = length ls" by (rule decompLength)
      from Some have zip: "pairs = zip ts ls" by (rule decompZip)
      with 3 Some True zip have "matcherL (set (zip ts ls @ ps)) = matcherLi (set qs)" by simp
      with len True show ?thesis by simp
    qed
  qed
next
  case (4 t x ps) thus ?case 
  proof (cases "matchList ps")
    case None with 4 show ?thesis by simp
  next
    case (Some pairs) with 4 show ?thesis by (auto simp: matcherLInsert)
  qed
qed

abbreviation rVars :: "(('f,'v)term \<times> ('f,'v)term)list \<Rightarrow> 'v set"
where "rVars ps \<equiv> \<Union> ((vars_term \<circ> snd) ` (set ps))"

abbreviation lVars :: "('v \<times> ('f,'v)term)list \<Rightarrow> 'v set"
where "lVars ps \<equiv> { x | t x. (x,t) \<in> set ps}"

lemma matchListVars : assumes "matchList ps = Some qs" shows "rVars ps = lVars qs"
using assms 
proof (induct arbitrary:qs rule:matchList.induct)
  case 1 thus ?case by auto
next
  case (2 x f ls ps) thus ?case by auto
next
  case (4 t x ps) thus ?case by (cases "matchList ps", auto)
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis 
    proof (cases "decomp ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some pairs)
      hence len: "length ts = length ls" by (rule decompLength)
      from Some have zip: "pairs = zip ts ls" by (rule decompZip)
      with 3 Some True zip have ind: "rVars (pairs @ ps) = lVars qs" by auto
      from len have "rVars (zip ts ls) = rVars [(Fun f ts, Fun g ls)]" 
      proof (induct ls arbitrary: ts, simp)
	case (Cons l ls)
	thus ?case by (cases ts, auto)
      qed
      with ind zip show "rVars ((Fun f ts, Fun g ls) # ps) = lVars qs" by auto
    qed
  qed
qed

fun matchVars :: "('v \<times> ('f,'v)term) list \<Rightarrow> ('f,'v)sub option"
  where "matchVars [] = Some \<iota>"
    |   "matchVars ((x,t) # ps) = (case lookup x ps of 
             None \<Rightarrow> (case matchVars ps of
                          Some \<sigma> \<Rightarrow> Some (Sub ((get_sub \<sigma>)(x := t)))
                        | None \<Rightarrow> None)
          | Some s \<Rightarrow> if t = s then matchVars ps else None)"


lemma matchVarsSound : assumes "matchVars ps = Some \<sigma>"
  shows "\<sigma> \<in> matcherLi (set ps)"
proof (cases \<sigma>)
  case (Sub sig)
  {
    assume "matchVars ps = Some (Sub sig)"
    hence "Sub sig \<in> matcherLi (set ps)" 
    proof (induct ps arbitrary: sig, simp add: matcherL_def)
      case (Cons xt qs)
      show ?case 
      proof (cases xt)
	case (Pair x t)
	show ?thesis 
	proof (cases "lookup x qs")
	  case (Some s)
	  show ?thesis 
	  proof (cases "s = t")
	    case False
	    with Some Pair Cons show ?thesis by simp
	  next
	    case True
	    from Some have "(x,s) \<in> set qs" by (rule lookup_sound)
	    with True Some Pair  have "set (Cons xt qs) = set qs" by auto
	    with Pair Cons True Some show ?thesis by auto
	  qed
	next
	  case None
	  with Cons Pair show ?thesis 	    
	  proof (cases "matchVars qs", simp)
	    case (Some \<delta>)
	    show ?thesis
	    proof (cases \<delta>)
	      case (Sub delta)
	      with Cons Pair None Some have id: "sig = delta(x := t)" by auto
	      with Cons Some Sub have ind: "Sub delta \<in> matcherLi (set qs)" by auto
	      from None have "\<not> (\<exists> t. (x,t) \<in> set qs)" by (rule lookup_complete)
	      with id ind Pair show ?thesis unfolding matcherL_def by auto
	    qed
	  qed
	qed
      qed
    qed		 
  }
  with assms Sub show ?thesis by auto
qed


lemma matchVarsComplete : assumes "\<delta> \<in> matcherLi (set ps)" 
  shows "\<exists> \<sigma>. matchVars ps = Some \<sigma> \<and> (\<forall> x \<in> lVars ps. get_sub \<delta> x = get_sub \<sigma> x)" 
proof (cases \<delta>)
  case (Sub del)
  {
    let ?prop = "\<lambda> sig ps. matchVars ps = Some (Sub sig) \<and> (\<forall> x \<in> lVars ps. del x = sig x)" 
    assume "Sub del \<in> matcherLi (set ps)"
    hence "\<exists> sig. ?prop sig ps" 
    proof (induct ps)
      case Nil
      hence "?prop (get_sub \<iota>) []" by (auto simp: sub_empty_def)
      thus ?case by blast
    next
      case (Cons xt qs)
      show ?case 
      proof (cases xt)
	case (Pair x t)
	with Cons have delt: "del x = t" unfolding matcherL_def by auto
	from Cons have "Sub del \<in> matcherLi (set qs)" unfolding matcherL_def by auto 
	with Cons obtain sig where ind: "?prop sig qs" by force
	show ?thesis 
	proof (cases "lookup x qs")
	  case (Some s)
	  hence  "(x,s) \<in> set qs" by (rule lookup_sound)
	  with Cons have dels: "del x = s" unfolding matcherL_def by auto
	  with ind `(x,s) \<in> set qs` have "del x = sig x" by auto
	  with dels delt ind Pair Some have "?prop sig (xt # qs)"  by auto 
	  thus ?thesis by blast
	next
	  case None
	  hence "\<not> (\<exists> s. (x,s) \<in> set qs)" by (rule lookup_complete)
	  have "?prop (sig( x:= t)) (xt # qs)" using ind Pair None delt by auto
	  thus ?thesis by blast
	qed
      qed
    qed
  }
  with Sub assms show ?thesis by auto
qed

fun match :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)sub option"
  where "match t l = (case matchList [(t,l)] of 
                          None \<Rightarrow> None 
                        | Some ps \<Rightarrow> matchVars ps)"


lemma matchComplete : assumes "l \<cdot> \<delta> = s" 
  shows "\<exists> \<sigma> . match s l = Some \<sigma> \<and> (\<forall> x \<in> vars_term l. get_sub \<delta> x = get_sub \<sigma> x)"
proof -
  from assms have match: "\<delta> \<in> matcherL {(s,l)}" unfolding matcherL_def by auto
  show ?thesis
  proof (cases "matchList [(s,l)]")
    case None
    hence "matcherL (set [(s,l)]) = {}" by (rule matchListComplete)
    with match show ?thesis by simp
  next
    case (Some ps)
    hence "matcherL (set [(s,l)]) = matcherLi (set ps)" by (rule matchListSound)    
    with match have "\<delta> \<in> matcherLi (set ps)" by auto
    hence "\<exists> \<sigma>. matchVars ps = Some \<sigma> \<and>  (\<forall>x\<in>lVars ps. get_sub \<delta> x = get_sub \<sigma> x)"
      by (rule matchVarsComplete)
    from this obtain \<sigma> where one: "matchVars ps = Some \<sigma> \<and>  (\<forall>x\<in>lVars ps. get_sub \<delta> x = get_sub \<sigma> x)" ..    
    from Some have "rVars [(s,l)] = lVars ps" by (rule matchListVars)    
    with Some one have "match s l = Some \<sigma> \<and> (\<forall>x\<in>vars_term l. get_sub \<delta> x = get_sub \<sigma> x)" by auto
    thus ?thesis ..
  qed
qed
      
lemma matchSound : assumes "match t l = Some \<sigma>" shows "l \<cdot> \<sigma> = t"
using assms
proof (cases "matchList [(t,l)]", simp)
  case (Some ps) note oSome = this
  thus ?thesis using assms
  proof (cases "matchVars ps", simp)
    case (Some \<delta>)
    with matchVarsSound have "\<delta> \<in> matcherLi (set ps)" by auto
    with matchListSound `matchList [(t,l)] = Some ps` have "\<delta> \<in> matcherL (set [(t,l)])" by force
    with Some oSome assms show ?thesis unfolding matcherL_def by simp
  qed
qed

declare match.simps[simp del]

definition matches :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "matches t l \<equiv> case match t l of None \<Rightarrow> False | Some _ \<Rightarrow> True"

lemma matches[simp]: "matches t l = (\<exists> \<sigma> . t = l \<cdot> \<sigma>)"
proof (cases "match t l")
  case (Some \<sigma>)
  with matchSound[OF Some] show ?thesis unfolding matches_def by auto
next
  case None
  have "\<not> (\<exists> \<sigma>. t = l \<cdot> \<sigma>)"
  proof (rule notI)
    assume "\<exists> \<sigma> . t = l \<cdot> \<sigma>"
    then obtain \<sigma> where t: "l \<cdot> \<sigma> = t" by auto
    from matchComplete[OF t] None show False by auto
  qed
  with None show ?thesis unfolding matches_def by auto
qed

(* checking \<exists> \<sigma> \<delta>. s \<sigma> = t \<delta> via mgu's and variable renamings *)
fun map_vars :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'w)term"
where "map_vars m (Var x) = (Var (m x))"
  | "map_vars m (Fun f ts) = Fun f (map (map_vars m) ts)"

definition mgu_var_disjoint :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,('v,'v)either)sub option"
where "mgu_var_disjoint s t = mgu (map_vars Left s) (map_vars Right t)"

fun merge_substs :: "('f,'v)sub \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,('v,'v)either) sub"
where "merge_substs (Sub sigma) (Sub delta) = Sub (\<lambda> x. (case x of Left y \<Rightarrow> map_vars Left (sigma y) | Right y \<Rightarrow> map_vars Left (delta y)))"

lemma merge_substs_left : "map_vars Left s \<cdot> (merge_substs \<sigma> \<delta>) = map_vars Left (s \<cdot> \<sigma>)"
proof (induct s)
  case (Var x)
  thus ?case by (cases \<sigma>, cases \<delta>, auto)
next
  case (Fun f ts)
  thus ?case by auto
qed

lemma merge_substs_right : "map_vars Right s \<cdot> (merge_substs \<sigma> \<delta>) = map_vars Left (s \<cdot> \<delta>)"
proof (induct s)
  case (Var x)
  thus ?case by (cases \<sigma>, cases \<delta>, auto)
next
  case (Fun f ts)
  thus ?case by auto
qed

lemma mgu_var_disjoint_complete: 
  fixes s t :: "('f,'v)term"
  assumes non_unif: "mgu_var_disjoint s t = None"
  shows "\<not> (\<exists> \<sigma> \<delta>. s \<cdot> \<sigma> = t \<cdot> \<delta>)" (is "\<not> ?E")
proof 
  assume ?E
  from this obtain \<sigma> \<delta> :: "('f,'v)sub" where id: "s \<cdot> \<sigma> = t \<cdot> \<delta>" by auto
  obtain \<gamma> where gamma: "\<gamma> = merge_substs \<sigma> \<delta>" by auto
  from id gamma merge_substs_right[of t] merge_substs_left[of s]
  have unif: "map_vars Left s \<cdot> \<gamma> = map_vars Right t \<cdot> \<gamma>" by auto
  with mguComplete[OF non_unif[unfolded mgu_var_disjoint_def]] 
  show False unfolding unifiers_def by auto
qed

fun left_subst :: "('f,('v,'v)either)sub \<Rightarrow> ('f,'v)sub"
where "left_subst (Sub sigma) = Sub (\<lambda> x. map_vars left (sigma (Left x)))"

fun right_subst :: "('f,('v,'v)either)sub \<Rightarrow> ('f,'v)sub"
where "right_subst (Sub sigma) = Sub (\<lambda> x. map_vars left (sigma (Right x)))"

lemma left_subst : "map_vars left (map_vars Left s \<cdot> \<gamma>) =  s \<cdot> (left_subst \<gamma>)"
proof (induct s)
  case (Fun f ts) thus ?case by auto
next
  case (Var x) thus ?case by (cases \<gamma>, auto)
qed

lemma map_vars_compose : "map_vars m1 (map_vars m2 t) = map_vars (m1 o m2) t" 
  by (induct t, auto)

lemma right_subst : "map_vars left (map_vars Right s \<cdot> \<gamma>) =  s \<cdot> (right_subst \<gamma>)"
proof (induct s)
  case (Fun f ts) thus ?case by auto
next
  case (Var x) thus ?case by (cases \<gamma>, auto)
qed

lemma mgu_var_disjoint_sound: 
  fixes s t :: "('f,'v)term"
  assumes unif: "mgu_var_disjoint s t = Some \<gamma>"
  shows "s \<cdot> (left_subst \<gamma>) = t \<cdot> (right_subst \<gamma>)"
proof -
  from mguSound[OF unif[unfolded mgu_var_disjoint_def], unfolded isSMgu_def unifiers_def]
  have "map_vars left (map_vars Left s \<cdot> \<gamma>) = map_vars left (map_vars Right t \<cdot> \<gamma>)" by auto
  thus ?thesis by (simp only: left_subst right_subst)
qed

(* if terms s and t can become identical via two substitutions \<sigma> and \<delta> 
   then mgu_var_disjoint yields a more general substitution \<mu> *)
lemma mgu_var_disjoint_instance:
  fixes s t :: "('f,'v)term"
  assumes unif_disj: "s \<cdot> \<sigma> = t \<cdot> \<delta>"
  shows "\<exists> \<mu>. mgu_var_disjoint s t = Some \<mu> \<and> 
  (\<forall> u. map_vars Left (u \<cdot> \<sigma>) = map_vars Left u \<cdot> \<mu> \<cdot> (merge_substs \<sigma> \<delta>)) \<and>
  (\<forall> u. map_vars Left (u \<cdot> \<delta>) = map_vars Right u \<cdot> \<mu> \<cdot> (merge_substs \<sigma> \<delta>))"
proof -
  obtain \<gamma> where gamma: "\<gamma> = merge_substs \<sigma> \<delta>" by auto
  from unif_disj gamma merge_substs_right[of t] merge_substs_left[of s]
  have unif: "map_vars Left s \<cdot> \<gamma> = map_vars Right t \<cdot> \<gamma>" (is "?s \<cdot> \<gamma> = ?t \<cdot> \<gamma>") by auto
  from mgu_of[OF unif] have unif2: "?s \<cdot> mgu_of ?s ?t = ?t \<cdot> mgu_of ?s ?t" and inst: "\<gamma> = mgu_of ?s ?t \<circ>s \<gamma>" by auto
  have "mgu_var_disjoint s t = Some (mgu_of ?s ?t)" unfolding mgu_var_disjoint_def mgu_of_def
    using mguComplete[unfolded unifiers_def] unif
    by (cases "mgu ?s ?t", auto)
  with inst obtain \<mu> where mu: "mgu_var_disjoint s t = Some \<mu>" and gamma_mu: "\<gamma> = \<mu> \<circ>s \<gamma>" by auto
  have inst_left: "\<forall> u. map_vars Left (u \<cdot> \<sigma>) = (map_vars Left u \<cdot> \<mu> \<cdot> \<gamma>)"
  proof
    fix u
    have "map_vars Left (u \<cdot> \<sigma>) = (map_vars Left u) \<cdot> \<gamma>" unfolding gamma by (simp only: merge_substs_left)
    also have "\<dots> = map_vars Left u \<cdot> \<mu> \<cdot> \<gamma>"
      by (simp only: sub_compose_apply_term_distrib[symmetric] gamma_mu[symmetric])
    finally show "map_vars Left (u \<cdot> \<sigma>) = map_vars Left u \<cdot> \<mu> \<cdot> \<gamma>" .
  qed
  have inst_right: "\<forall> u. map_vars Left (u \<cdot> \<delta>) = (map_vars Right u \<cdot> \<mu> \<cdot> \<gamma>)"
  proof
    fix u
    have "map_vars Left (u \<cdot> \<delta>) = (map_vars Right u) \<cdot> \<gamma>" unfolding gamma by (simp only: merge_substs_right)
    also have "\<dots> = map_vars Right u \<cdot> \<mu> \<cdot> \<gamma>" by (simp only: sub_compose_apply_term_distrib[symmetric] gamma_mu[symmetric])
    finally show "map_vars Left (u \<cdot> \<delta>) = map_vars Right u \<cdot> \<mu> \<cdot> \<gamma>" .
  qed
  from mu gamma inst_left inst_right show ?thesis by auto
qed

end
